from __future__ import annotations

from abc import ABC

import numpy as np

from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import full_range, pi, xexpexp1
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveVariable
from osaft.solutions.doinikov1994rigid.arf_limiting import ARFLimiting


class ARFArbitraryBL(ARFLimiting, ABC):
    """ARF according to Doinikov's theory (viscous fluid-rigid sphere; 1994)

    .. note::
        This classes implements all attributes and methods to compute the ARF
        for the case where both particle radius and boundary layer thickness
        are small compared to the wavelength
        :math:`|x| \\ll 1`, :math:`1 \\ll |x_v|`.
        This class does not actually implement a method for the ARF. If you
        want to compute the ARF use
        :attr:`~osaft.solutions.doinikov1994rigid.ARF`

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param position: Position in the standing wave field [rad]
    :param long_wavelength: using :math:`x \\ll 1`
    :param small_boundary_layer: :math:`x \\ll x_v \\ll 1`
    :param large_boundary_layer: :math`x \\ll 1 \\ll x_v`
    :param fastened_sphere: use theory of fastened sphere
    :param background_streaming: background streaming contribution
    :param N_max: Highest order mode
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float,
        long_wavelength: bool,
        small_boundary_layer: bool,
        large_boundary_layer: bool,
        fastened_sphere: bool,
        background_streaming: bool,
        N_max: int,
    ) -> None:
        """Constructor method"""

        # Call to init method of parent class
        super().__init__(
            f,
            R_0,
            rho_s,
            rho_f,
            c_f,
            eta_f,
            zeta_f,
            p_0,
            wave_type,
            position,
            long_wavelength,
            small_boundary_layer,
            large_boundary_layer,
            fastened_sphere,
            background_streaming,
            N_max,
        )

        # Coefficients G_n
        self._G_0 = ActiveVariable(self._compute_G_0, "G_0")
        self._G_1 = ActiveVariable(self._compute_G_1, "G_1")
        self._G_2 = ActiveVariable(self._compute_G_2, "G_2")
        self._G_3 = ActiveVariable(self._compute_G_3, "G_3")
        self._G_4 = ActiveVariable(self._compute_G_4, "G_4")

        self._G_0.is_computed_by(self._rho_t, self._x_v)
        self._G_1.is_computed_by(self._x_v)
        self._G_2.is_computed_by(self._x_v)
        self._G_3.is_computed_by(self._x_v)
        self._G_4.is_computed_by(self._x_v)

        # Coefficients D_n
        self._D_0 = ActiveVariable(self._compute_D_0, "D_0")
        self._D_1 = ActiveVariable(self._compute_D_1, "D_1")

        self._D_0.is_computed_by(self._x, self._x_v, self._rho_t)
        self._D_1.is_computed_by(self._x, self._x_v, self._rho_t)

    def _F_1_limit(self) -> float:
        def term(n: int) -> float:
            A_n = self.A_in(n)
            A_n1 = self.A_in(n + 1)
            out = self.D_n_limit(n) * A_n * A_n1.conjugate()
            out += self.D_n_limit(n).conjugate() * A_n.conjugate() * A_n1
            out *= (n + 1) / (2 * n + 1) / (2 * n + 3)
            return out.real

        arf = np.sum([term(n) for n in full_range(1)])
        return -3 / 2 * pi * self.rho_f * arf

    # -------------------------------------------------------------------------
    # Small Particle ARF
    # -------------------------------------------------------------------------

    def _arf_small_particle(self) -> float:
        """Acoustic radiation force for the case where the particle radius
        and the viscous boundary layer are small compared to the wavelength
        """
        if self.wave_type == WaveType.TRAVELLING:
            return self._arf_small_particle_travelling()
        else:
            return self._arf_small_particle_standing()

    def _arf_small_particle_travelling(self) -> float:
        """Eq. (5.10), (5.12), (6.1), (6.2)"""
        # F1
        F1 = self._F_1_limit()
        # F_2
        F2 = self._F_2_travelling() if self.background_streaming else 0
        # S_9n contribution to background streaming
        S_9n = 0 if self.background_streaming else self.S_9n_contribution()
        return F1 + F2 - S_9n

    def _arf_small_particle_standing(self) -> float:
        """Eq. (5.15), (5.16), (6.1), (6.2)"""
        # F1
        F1 = self._F_1_limit()
        # F_2
        F2 = self._F_2_standing() if self.background_streaming else 0
        # S_9n contribution to background streaming
        S_9n = 0 if self.background_streaming else self.S_9n_contribution()
        return F1 + F2 - S_9n

    # -------------------------------------------------------------------------
    # Coefficients D_n
    # -------------------------------------------------------------------------

    def D_n_limit(self, n: int) -> complex:
        """Approximation for the coefficient D_0, D_1 from Eq (6.1), (6.2)"""
        if n == 0:
            return self.D_0
        elif n == 1:
            return self.D_1
        else:
            raise ValueError("D_n in the limit is only defined for n = 0, 1")

    @property
    def D_0(self) -> complex:
        """Approximation for the coefficient D_0 from Eq (6.1)"""
        return self._D_0.value

    def _compute_D_0(self) -> complex:
        out = self.G_1
        out += self.x_v**3 * (12 + self.x_v**2) * self._f(self.x_v)
        out *= -self.x**3 * self.G_0
        out += 2 * self.x**3 / 9
        out += self.x**3 / (3 * self.x_v**2)
        return out

    @property
    def D_1(self) -> complex:
        """Approximation for the coefficient D_1 from Eq (6.2)"""
        return self._D_1.value

    def _compute_D_1(self):
        out = -2 * (1 + 1j) * (9 + self.x_v**2)
        out *= self._f(self.x_v - 1j * self.x_v)
        out += 2 * self.G_3 * self._f(self.x_v)
        out += 1j * self.G_4 * self._f(-1j * self.x_v)
        out *= -self.x_v**3
        if abs(out.real + self.G_2.real) / abs(self.G_2.real) < 1e-6:
            RuntimeWarning(  # pragma: no cover
                "Possible precision loss detected. "
                "Consider using a limiting case solution.",
            )
        out += self.G_2
        out *= self.x**3 * self.G_0.conjugate() / (1 + self.x_v)
        out += self.x**3 / (3 * self.x_v**2)
        return out

    @staticmethod
    def _f(z):
        return xexpexp1(z)

    # -------------------------------------------------------------------------
    # Coefficients S_9n
    # -------------------------------------------------------------------------

    def S_9n_contribution(self) -> complex:
        """S_9n contribution to the background streaming"""

        # S_90 + #S_91 (background streaming contribution)
        S_9n = 2 * self.x**3 * self.x_v ** (-2) / 9

        term_1 = S_9n * self.A_in(0) * self.A_in(1).conjugate()
        term_1 += S_9n.conjugate() * self.A_in(1) * self.A_in(0).conjugate()
        term_2 = S_9n * self.A_in(1) * self.A_in(2).conjugate()
        term_2 += S_9n.conjugate() * self.A_in(2) * self.A_in(1).conjugate()

        return (2 * term_1 / 15 + term_2 / 3).real

    # -------------------------------------------------------------------------
    # Coefficients G_n
    # -------------------------------------------------------------------------

    @property
    def G_0(self) -> complex:
        """Approximation to value G_0 from Eq (6.3)"""
        return self._G_0.value

    def _compute_G_0(self) -> complex:
        denominator = 9 * self.rho_t + 9 * self.rho_t * self.x_v
        denominator += (2 + self.rho_t) * self.x_v**2
        out = (1 - self.rho_t) / (72 * denominator)
        return out

    @property
    def G_1(self) -> complex:
        """Approximation to value G_1 from Eq (6.4)"""
        return self._G_1.value

    def _compute_G_1(self) -> complex:
        out = 48 - 96 * self.x_v + 2 * self.x_v**2 - 14 * self.x_v**3
        out += self.x_v**4 - self.x_v**5
        return out

    @property
    def G_2(self) -> complex:
        """Approximation to value G_2 from Eq (6.5)"""
        return self._G_2.value

    def _compute_G_2(self) -> complex:
        out = 48 + (48 + 192 * 1j / 5) * self.x_v
        out += (122 + 192 * 1j) * self.x_v**2 / 5
        out += 42 * (1 - 1j) * self.x_v**3 / 5
        out += (49 + 36 * 1j) * self.x_v**4 / 10
        out -= (31 - 17 * 1j) * self.x_v**5 / 10
        out += (6 + 31 * 1j) * self.x_v**6 / 30
        out += (1 + 6 * 1j) * self.x_v**7 / 30
        out += 1j * self.x_v**8 / 30
        return out

    @property
    def G_3(self) -> complex:
        """Approximation to value G_3 from Eq (6.6)"""
        return self._G_3.value

    def _compute_G_3(self) -> complex:
        return 3 - 3 * 1j * self.x_v - self.x_v**2

    @property
    def G_4(self) -> complex:
        """Approximation to value G_4 from Eq (6.7)"""
        return self._G_4.value

    def _compute_G_4(self) -> complex:
        out = 9 + 9 * self.x_v + 41 * self.x_v**2 / 10
        out += 11 * self.x_v**3 / 10 + self.x_v**4 / 5 + self.x_v**5 / 30
        return out


if __name__ == "__main__":
    pass



from osaft import WaveType
from osaft.solutions import (
    doinikov1994compressible,
    doinikov1994rigid,
    doinikov2021viscoelastic,
    doinikov2021viscous,
    gorkov1962,
    hasegawa1969,
    king1934,
    settnes2012,
    yosioka1955,
)
from osaft.tests.testing_parameters import RandomParameters


class SolutionFactory:

    parameters = RandomParameters()

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def king_1934_base(self, new: bool = False) -> king1934.BaseKing:
        """Returns an initiated class of ``king1934.BaseKing``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()
        sol = king1934.BaseKing(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def king_1934_scattering(
        self,
        new: bool = False,
    ) -> king1934.ScatteringField:
        """Returns an initiated class of ``king1934.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()
        sol = king1934.ScatteringField(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def king_1934_arf(self, new: bool = False) -> king1934.ARF:
        """Returns an initiated class of ``king1934.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()
        sol = king1934.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def gorkov_1962_arf(self, new: bool = False) -> gorkov1962.ARF:
        """Returns an initiated class of ``gorkov1962.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = gorkov1962.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.parameters.c_s,
            self.rho_f,
            self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def yosioka_1955_base(
        self,
        new: bool = False,
    ) -> yosioka1955.BaseYosioka:
        """Returns an initiated class of ``yosioka1955.BaseYosioka``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = yosioka1955.BaseYosioka(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            c_s=self.c_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )

        return sol

    def yosioka_1955_scattering(
        self,
        new: bool = False,
    ) -> yosioka1955.ScatteringField:
        """Returns an initiated class of ``yosioka1955.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = yosioka1955.ScatteringField(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            c_s=self.c_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )

        return sol

    def yosioka_1955_arf(
        self,
        new: bool = False,
    ) -> yosioka1955.ARF:
        """Returns an initiated class of ``yosioka1955.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = yosioka1955.ARF(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            c_s=self.c_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )

        return sol

    def hasegawa_1969_base(
        self,
        new: bool = False,
    ) -> hasegawa1969.BaseHasegawa:
        """Returns an initiated class of ``hasegawa1969.BaseHasegawa``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = hasegawa1969.BaseHasegawa(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def hasegawa_1969_scattering(
        self,
        new: bool = False,
    ) -> hasegawa1969.ScatteringField:
        """Returns an initiated class of ``hasegawa1969.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = hasegawa1969.ScatteringField(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def hasegawa_1969_arf(
        self,
        new: bool = False,
    ) -> hasegawa1969.ARF:
        """Returns an initiated class of ``hasegawa1969.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = hasegawa1969.ARF(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def doinikov_1994_rigid_scattering(
        self,
        new: bool = False,
    ) -> doinikov1994rigid.ScatteringField:
        """Returns an initiated class of ``doinikov1994rigid.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov1994rigid.ScatteringField(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def doinikov_1994_rigid_arf(
        self,
        new: bool = False,
    ) -> doinikov1994rigid.ARF:
        """Returns an initiated class of ``doinikov1994rigid.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov1994rigid.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def doinikov_1994_compressible_scattering(
        self,
        new: bool = False,
    ) -> doinikov1994compressible.ScatteringField:
        """Returns an initiated class of
        ``doinikov1994compressible.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov1994compressible.ScatteringField(
            self.f,
            self.R_0,
            self.rho_s,
            self.c_s,
            self.eta_s,
            self.zeta_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def doinikov_1994_compressible_arf(
        self,
        new: bool = False,
    ) -> doinikov1994compressible.ARF:
        """Returns an initiated class of ``doinikov1994compressible.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov1994compressible.ARF(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            c_s=self.c_s,
            eta_s=self.eta_s,
            zeta_s=self.zeta_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            eta_f=self.eta_f,
            zeta_f=self.zeta_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def settnes_2012_arf(
        self,
        new: bool = False,
    ) -> settnes2012.ARF:
        """Returns an initiated class of ``settnes2021.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = settnes2012.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.c_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        return sol

    def doinikov_2021_viscous_scattering(
        self,
        new: bool = False,
    ) -> doinikov2021viscous.ScatteringField:
        """Returns an initiated class of
        ``doinikov2021viscous.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov2021viscous.ScatteringField(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            eta_f=self.eta_f,
            zeta_f=self.zeta_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def doinikov_2021_viscous_streaming(
        self,
        new: bool = False,
    ) -> doinikov2021viscous.StreamingField:
        """Returns an initiated class of ``doinikov2021viscous.StreamingField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov2021viscous.StreamingField(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            eta_f=self.eta_f,
            zeta_f=self.zeta_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def doinikov_2021_viscous_arf(
        self,
        new: bool = False,
    ) -> doinikov2021viscous.ARF:
        """Returns an initiated class of ``doinikov2021viscous.ARF``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov2021viscous.ARF(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            eta_f=self.eta_f,
            zeta_f=self.zeta_f,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    def doinikov_2021_viscoelastic_scattering(
        self,
        new: bool = False,
    ) -> doinikov2021viscoelastic.ScatteringField:
        """Returns an initiated class of
        ``doinikov2021viscoelastic.ScatteringField``

        :param new: if ``True`` new random parameter values are computed
        """
        if new:
            self.parameters.change_all()

        sol = doinikov2021viscoelastic.ScatteringField(
            f=self.f,
            R_0=self.R_0,
            rho_s=self.rho_s,
            E_s=self.E_s,
            nu_s=self.nu_s,
            rho_f=self.rho_f,
            c_f=self.c_f,
            eta_f=self.eta_f,
            zeta_f=self.zeta_f,
            eta_p=self.eta_p,
            zeta_p=self.eta_p,
            lambda_M=self.lambda_M,
            p_0=self.p_0,
            wave_type=self.wave_type,
            position=self.position,
        )
        return sol

    # -------------------------------------------------------------------------
    # Getters & Setters
    # -------------------------------------------------------------------------

    @property
    def R_0(self) -> float:
        return self.parameters.R_0

    @property
    def f(self) -> float:
        return self.parameters.f

    @property
    def c_f(self) -> float:
        return self.parameters.c_f

    @property
    def rho_f(self) -> float:
        return self.parameters.rho_f

    @property
    def c_s(self) -> float:
        return self.parameters.c_s

    @property
    def eta_s(self) -> float:
        return self.parameters.eta_s

    @property
    def zeta_s(self) -> float:
        return self.parameters.zeta_s

    @property
    def rho_s(self) -> float:
        return self.parameters.rho_s

    @property
    def eta_f(self) -> float:
        return self.parameters.eta_f

    @property
    def zeta_f(self) -> float:
        return self.parameters.zeta_f

    @property
    def eta_p(self) -> float:
        return self.parameters.eta_p

    @property
    def zeta_p(self) -> float:
        return self.parameters.zeta_p

    @property
    def lambda_M(self) -> float:
        return self.parameters.lambda_M

    @property
    def E_s(self) -> float:
        return self.parameters.E_s

    @property
    def nu_s(self) -> float:
        return self.parameters.nu_s

    @property
    def p_0(self) -> float:
        return self.parameters.p_0

    @property
    def position(self) -> float:
        return self.parameters.position

    @property
    def wave_type(self) -> WaveType:
        return self.parameters.wave_type

    @property
    def N_max(self):
        return self.parameters.N_max


if __name__ == "__main__":
    pass



import unittest

from osaft.core.functions import BesselFunctions as Bf
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.basetest_scattering import HelperScattering
from osaft.tests.solution_factory import SolutionFactory


class TestScattering(HelperScattering, BaseTestSolutions):
    def setUp(self) -> None:

        super().setUp()

        self._v_particle_threshold = 1e-5
        self._v_fluid_threshold = 1e-3
        self._v_boundary_conditions = 1e-8

        self.cls = SolutionFactory().doinikov_1994_compressible_scattering()

    # ------------------
    # fluid velocity
    # ------------------

    def V_r_sc(self, n, r) -> complex:
        # grad(phi)
        out = self.cls.k_f * self.alpha_n(n)
        out *= Bf.d1_hankelh1(n, self.cls.k_f * r)
        # curl(Psi)
        arg = self.cls.k_v * r
        out -= n * (n + 1) / r * self.beta_n(n) * Bf.hankelh1(n, arg)
        out *= self.cls.A_in(n)
        return out

    def V_theta_sc(self, n, r) -> complex:
        # curl(Psi)
        arg_v = self.cls.k_v * r
        out = Bf.hankelh1(n, arg_v)
        out += arg_v * Bf.d1_hankelh1(n, arg_v)
        out *= -self.beta_n(n)
        # grad(phi)
        out += self.alpha_n(n) * Bf.hankelh1(n, self.cls.k_f * r)

        out *= self.cls.A_in(n) / r
        return out

    # ------------------
    # particle velocity
    # ------------------

    def radial_particle_velocity(self, r, theta, t, mode):
        def vel(n, r):
            out = self.cls.k_s * self.alpha_hat_n(n)
            out *= Bf.d1_besselj(n, r * self.cls.k_s)
            arg = r * self.cls.k_vs
            out -= n * (n + 1) / r * self.beta_hat_n(n) * Bf.besselj(n, arg)
            out *= self.cls.A_in(n)
            return out

        out = self.cls.radial_mode_superposition(
            vel,
            r,
            theta,
            t,
            mode,
        )

        return out

    def tangential_particle_velocity(self, r, theta, t, mode):
        def vel(n, r):
            arg = r * self.cls.k_vs
            out = Bf.besselj(n, arg) + arg * Bf.d1_besselj(n, arg)
            out *= -self.beta_hat_n(n)
            out += self.alpha_hat_n(n) * Bf.besselj(n, r * self.cls.k_s)
            out *= self.cls.A_in(n) / r
            return out

        out = self.cls.tangential_mode_superposition(
            vel,
            r,
            theta,
            t,
            mode,
        )

        return out

    # ------------------
    # generic test method for coefficients
    # ------------------

    def test_coefficients(self) -> None:
        dict_of_methods = {}
        names = [
            "alpha_n",
            "beta_n",
            "alpha_hat_n",
            "beta_hat_n",
            "potential_coefficient",
        ]
        for name in names:
            dict_of_methods[name] = None

        self._test_methods_n(dict_of_methods, threshold=1e-8)

    # ------------------
    # testing scattering coefficients
    # ------------------

    def alpha_n(self, n: int) -> complex:
        return self.cls.det_M_n(n, 0) / self.cls.det_M_n(n)

    def beta_n(self, n: int) -> complex:
        if n > 0:
            return self.cls.det_M_n(n, 1) / self.cls.det_M_n(n)
        else:
            return 0

    def alpha_hat_n(self, n: int) -> complex:
        return self.cls.det_M_n(n, 2) / self.cls.det_M_n(n)

    def beta_hat_n(self, n: int) -> complex:
        if n > 0:
            return self.cls.det_M_n(n, 3) / self.cls.det_M_n(n)
        else:
            return 0

    def potential_coefficient(self, n: int) -> complex:
        return self.cls.alpha_n(n) * self.cls.field.A_in(n)

    # ------------------
    @property
    def x(self):
        return self.cls.x

    @property
    def x_v(self):
        return self.cls.x_v

    @property
    def x_hat(self):
        return self._compute_x_hat()

    def _compute_x_hat(self) -> complex:
        return self.R_0 * self.cls.k_s

    def test_x(self) -> None:
        self.do_testing(
            func_1=self._compute_x_hat,
            func_2=lambda: self.cls.x_hat,
        )

    # ------------------

    @property
    def x_hat_v(self):
        return self._compute_x_hat_v()

    def _compute_x_hat_v(self) -> complex:
        return self.R_0 * self.cls.k_vs

    def test_x_hat_v(self) -> None:
        self.do_testing(
            func_1=self._compute_x_hat_v,
            func_2=lambda: self.cls.x_hat_v,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from osaft import gorkov1962, settnes2012
from osaft.core.backgroundfields import WaveType, WrongWaveTypeError
from osaft.core.functions import pi, sin
from osaft.tests.basetest_arf import (
    HelperCompareARF,
    HelperStandingARF,
    HelperTravelingARF,
)
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestSettnes(BaseTestSolutions):
    def setUp(self) -> None:

        super().setUp()

        self.cls = SolutionFactory().settnes_2012_arf()

        self.list_cls = [self.cls]

    # ------------------------------------------------------------------------
    # Methods
    # ------------------------------------------------------------------------

    @property
    def f_1(self) -> float:
        return 1 - self.cls.kappa_s / self.cls.kappa_f

    @property
    def f_2(self) -> complex:
        return (
            2
            * (1 - self.gamma)
            * (self.rho_t - 1)
            / (2 * self.rho_t + 1 - 3 * self.gamma)
        )

    @property
    def delta_t(self) -> float:
        return self.cls.delta / self.R_0

    @property
    def gamma(self) -> complex:
        return -3 / 2 * (1 + 1j * (1 + self.delta_t)) * self.delta_t

    @property
    def kappa_t(self) -> float:
        return self.cls.kappa_s / self.cls.kappa_f

    @property
    def rho_t(self) -> float:
        return self.rho_s / self.rho_f

    @property
    def Phi(self) -> float:
        return self.f_1 / 3 + self.f_2.real / 2

    # ------------------------------------------------------------------------
    # Tests
    # ------------------------------------------------------------------------

    def test_properties(self) -> None:
        properties = [
            "f_1",
            "f_2",
            "delta_t",
            "gamma",
            "kappa_t",
            "rho_t",
            "Phi",
        ]
        self._test_properties(properties)

    def test_wrong_wavetype(self):
        self.cls.wave_type = "wrong_wave_type"
        self.assertRaises(
            WrongWaveTypeError,
            self.cls.compute_arf,
        )

    def test_not_small_particle(self):
        # This method is not actually testing, only for coverage
        self.cls.R_0 = self.cls.fluid.lambda_f
        self.cls.compute_arf()


class TestCompareToGorkov(BaseTestSolutions, HelperCompareARF):
    def setUp(self):
        super().setUp()

        self.arf_compare_threshold = 1e-4
        self.parameters.eta_f = 0
        self.parameters.list_parameters.remove(self.parameters._eta_f)
        self.parameters._wave_type.list_of_values = [WaveType.STANDING]
        self.parameters.wave_type = WaveType.STANDING

        self.cls = settnes2012.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.c_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )

        self.compare_cls = gorkov1962.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.c_s,
            self.rho_f,
            self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
        )

        self.list_cls = [self.cls, self.compare_cls]


class TestStanding(TestSettnes, HelperStandingARF):
    def compute_arf(self) -> float:
        return (
            4
            * pi
            * self.Phi
            * self.R_0**3
            * self.cls.k_f.real
            * self.cls.E_ac
            * sin(2 * self.position)
        )


class TestTraveling(TestSettnes, HelperTravelingARF):
    def compute_arf(self) -> float:
        return (
            2
            * pi
            * self.R_0**3
            * self.cls.k_f.real
            * self.f_2.imag
            * self.cls.E_ac
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from osaft.core.functions import pi, sin, xexpexp1
from osaft.tests.basetest_arf import HelperStandingARF, HelperTravelingARF
from osaft.tests.solutions.doinikov1994rigid.setup_test_arf import BaseTestARF


class TestARFArbitraryBL(BaseTestARF):

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def x(self) -> complex:
        return self.cls.x

    @property
    def x_v(self) -> complex:
        return self.cls.x_v

    @property
    def norm_delta(self) -> complex:
        return self.cls.delta / self.R_0

    # -------------------------------------------------------------------------
    # D_n
    # -------------------------------------------------------------------------

    def D_n_limit(self, n: int) -> complex:
        if n == 0:
            return self.D_0
        elif n == 1:
            return self.D_1
        else:
            raise ValueError("D_n in the limit is only defined for n = 0, 1")

    @property
    def D_0(self):
        out = self.G_1
        out += self.x_v**3 * (12 + self.x_v**2) * self.function(self.x_v)
        out *= -self.x**3 * self.G_0
        out += 2 * self.x**3 / 9
        out += self.x**3 / (3 * self.x_v**2)
        return out

    @property
    def D_1(self):
        out = -2 * (1 + 1j) * (9 + self.x_v**2)
        out *= self.function(self.x_v - 1j * self.x_v)
        out += 2 * self.G_3 * self.function(self.x_v)
        out += 1j * self.G_4 * self.function(-1j * self.x_v)
        out *= -self.x_v**3
        out += self.G_2
        out *= self.cls.x**3 * self.G_0.conjugate() / (1 + self.x_v)
        out += self.cls.x**3 / (3 * self.x_v**2)
        return out

    @staticmethod
    def function(z):
        return xexpexp1(z)

    # -------------------------------------------------------------------------
    # G_n
    # -------------------------------------------------------------------------

    @property
    def G_0(self) -> complex:
        denominator = 9 * self.rho_t + 9 * self.rho_t * self.x_v
        denominator += (2 + self.rho_t) * self.x_v**2
        denominator *= 72

        numerator = 1 - self.rho_t
        return numerator / denominator

    @property
    def G_1(self) -> complex:
        out = 48 - 96 * self.x_v + 2 * self.x_v**2 - 14 * self.x_v**3
        out += self.x_v**4 - self.x_v**5
        return out

    @property
    def G_2(self):
        out = 48 + (48 + 192 * 1j / 5) * self.x_v
        out += (122 + 192 * 1j) * self.x_v**2 / 5
        out += 42 * (1 - 1j) * self.x_v**3 / 5
        out += (49 + 36 * 1j) * self.x_v**4 / 10
        out -= (31 - 17 * 1j) * self.x_v**5 / 10
        out += (6 + 31 * 1j) * self.x_v**6 / 30
        out += (1 + 6 * 1j) * self.x_v**7 / 30
        out += 1j * self.x_v**8 / 30
        return out

    @property
    def G_3(self):
        return 3 - 3 * 1j * self.x_v - self.x_v**2

    @property
    def G_4(self):
        out = 9 + 9 * self.x_v + 41 * self.x_v**2 / 10
        out += 11 * self.x_v**3 / 10
        out += self.x_v**4 / 5 + self.x_v**5 / 30
        return out

    # -------------------------------------------------------------------------
    # F2 (Streaming contribution)
    # -------------------------------------------------------------------------

    def F_2_standing(self):
        k = self.cls.k_f
        x = self.x
        x_v = self.x_v
        d = self.cls.abs_pos
        first = sin(k * d + k.conjugate() * d) * sin(x + x.conjugate())
        first *= (x - x.conjugate()) / (x + x.conjugate())
        second = sin(k * d - k.conjugate() * d) * sin(x - x.conjugate())
        second *= (x + x.conjugate()) / (x - x.conjugate())
        out = first - second
        out *= -3 / 4 * pi * self.rho_f * self.cls.field.abs_A_squared
        out *= x * x.conjugate()
        out /= x_v**2
        return out

    def F_2_travelling(self):
        x = self.x
        x_v = self.x_v

        out = sin(x - x.conjugate()) / (x - x.conjugate()) * 1j / x_v**2
        out *= x * x.conjugate() * (x + x.conjugate())
        out *= -3 / 2 * pi * self.rho_f * self.cls.field.abs_A_squared
        return out

    def test_F_2_standing(self):
        self.do_testing(
            self.cls._F_2_standing,
            self.F_2_standing,
            threshold=1e-8,
        )

    def test_F_2_travelling(self):
        self.do_testing(
            self.cls._F_2_travelling,
            self.F_2_travelling,
            threshold=1e-8,
        )

    def test_methods_n(self):

        dict_methods = dict()
        dict_methods["D_n_limit"] = None
        self.background_streaming = True
        self.assign_parameters()
        self._test_methods_n(dict_methods, n_end=1, threshold=1e-8)

    def test_properties(self):
        dict_properties = dict()
        dict_properties["G_0"] = None
        dict_properties["G_1"] = None
        dict_properties["G_2"] = None
        dict_properties["G_4"] = None
        dict_properties["D_0"] = None
        dict_properties["D_1"] = None
        self.background_streaming = True
        self.assign_parameters()
        self._test_properties(dict_properties, threshold=1e-6)

    def test_D_n_error(self):
        self.assertRaises(ValueError, self.cls.D_n_limit, 2)


# |x| << 1  |x| << |x_v|
class TestSmallParticleTravelling(BaseTestARF, HelperTravelingARF):
    def setUp(self):
        super().setUp()
        self._threshold = 1e-3
        self.long_wavelength = True
        self.fastened_sphere = False
        self.large_boundary_layer = False
        self.small_boundary_layer = False
        self.background_streaming = True
        self.assign_parameters()

    def compute_arf(self):
        D_0 = self.cls.D_0
        D_1 = self.cls.D_1
        F1 = D_0 - D_0.conjugate()
        F1 += 2 * (D_1 - D_1.conjugate())
        F1 *= 3 / 2 * 1j * pi * self.rho_f * self.cls.field.abs_A_squared
        F2 = self.cls._F_2_travelling() if self.background_streaming else 0
        S_9 = 0 if self.background_streaming else self.cls.S_9n_contribution()
        return (F1 + F2 - S_9).real


class TestSmallParticleStanding(BaseTestARF, HelperStandingARF):
    def setUp(self):
        super().setUp()
        self._threshold = 1e-3
        self.long_wavelength = True
        self.fastened_sphere = False
        self.large_boundary_layer = False
        self.small_boundary_layer = False
        self.background_streaming = True
        self.assign_parameters()

    def compute_arf(self):
        D_0 = self.cls.D_0
        D_1 = self.cls.D_1
        sine = sin(2 * self.cls.k_f * self.cls.abs_pos)
        sine_conj = sin(2 * self.cls.k_f.conjugate() * self.cls.abs_pos)
        F1 = D_0 * sine + D_0.conjugate() * sine_conj
        F1 -= 2 * (D_1 * sine + D_1.conjugate() * sine_conj)
        F1 *= 3 / 4 * pi * self.rho_f * self.cls.field.abs_A_squared
        F2 = self.cls._F_2_standing() if self.background_streaming else 0
        S_9 = 0 if self.background_streaming else self.cls.S_9n_contribution()
        return (F1 + F2 - S_9).real


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import collections
from collections.abc import Callable, Sequence
from numbers import Number

import matplotlib.pyplot as plt
import numpy as np

from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.plotting.arf.arf_plotter import ARFPlotter
from osaft.plotting.datacontainers.arf_datacontainer import ARFData
from osaft.solutions.base_arf import BaseARF

NDArray = np.ndarray


class ARFPlot:
    """Plotting different ARF solutions inside same axis over attribute

    :param attr_name: name of attribute for x-axis
    :param x_values: x values for axis
    """

    def __init__(
        self,
        attr_name: None | str = None,
        x_values: None | NDArray = None,
    ):
        """Constructor method"""
        self.plotter = ARFPlotter()
        self._solutions = {}

        self._attr_name = PassiveVariable(attr_name, "Name of attribute")
        self._x_values = PassiveVariable(x_values, "Name of attribute")

        # the callable is not important; we want to use the dependency logic of
        # the ActiveVariable and we are not actually interested in the value
        self._needs_computation = ActiveVariable(
            lambda: None,
            "ARF needs to be recomputed",
        )

        self._needs_computation.is_computed_by(self._attr_name, self._x_values)

    @property
    def attr_name(self) -> str:
        """Attribute that is used as x-axis

        :getter: returns the attribute name for the x-axis
        :setter: sets attribute name
        """
        return self._attr_name.value

    @attr_name.setter
    def attr_name(self, value: str) -> None:
        self._attr_name.value = value

    @property
    def x_values(self) -> NDArray:
        """Values of x-axis

        :getter: returns x-axis values
        :setter: sets x-axis values
        """
        return self._x_values.value

    @x_values.setter
    def x_values(self, values: NDArray) -> None:
        self._x_values.value = values

    def set_abscissa(self, x_values: NDArray, attr_name: str) -> None:
        """Setting the abscissa variable and values for the ARF plot

        :param x_values: data points to be plotted over
        :param attr_name: name of the dependent variable to be plotted over
        """
        self.x_values = x_values
        self.attr_name = attr_name

    def add_solutions(
        self, *solutions: BaseARF, multicore: bool = False
    ) -> None:
        """Add solutions to list of solutions for plotting e.g.
        :class:`osaft.king1934.ARF()`

        :param solutions: one or multiple solutions,
               e.g. :class:`osaft.king1934.ARF()`
        :param multicore: it ``True``, multiple processing
        """
        if len(solutions) == 0:
            out = "is_computed_by() takes at least one positional argument "
            out += "(0 were given)"
            raise TypeError(out)
        for solution in solutions:
            if solution.name in self._solutions:
                raise ValueError(
                    f"Solution with the same name ({solution.name}) is "
                    "already in the list of solution and has been "
                    "overwritten. Consider renaming the "
                    f"attribute `name` of this solution({solution}).",
                )
            self._solutions[solution.name] = ARFData(
                solution,
                multicore=multicore,
            )

    def remove_solution(self, solution: BaseARF) -> None:
        """Remove solution of list of solutions for plotting

        :param solution: specific ARF solution, e.g.
            :class:`osaft.king1932.ARF()`
        """
        self._solutions.pop(solution.name, None)

    def _compute_arf(self) -> None:
        if not self._needs_computation.needs_update:
            return

        # value of the ActiveVariable needs to be accessed
        # such that the needs_update property changes to False
        _ = self._needs_computation.value
        for _, items in self._solutions.items():
            items.compute_arf(self.attr_name, self.x_values)

    def _find_max(self):
        """Finds the max ARF value in all solutions."""
        all_max = 0
        for _, item in self._solutions.items():
            current_max = np.max(abs(item._arf))
            if current_max > all_max:
                all_max = current_max
        return all_max

    def _normalize_arf(
        self,
        normalization: str | Number | Sequence | Callable,
    ) -> None:
        if normalization is None:
            return
        elif isinstance(normalization, str) and normalization == "max":
            norm = self._find_max()
        elif (
            isinstance(normalization, str)
            and normalization in self._solutions.keys()
        ):
            norm = self._solutions[normalization].arf
        elif isinstance(normalization, Number):
            norm = normalization
        elif (
            isinstance(normalization, collections.abc.Sequence)
            or isinstance(normalization, NDArray)
        ) and not isinstance(normalization, str):
            norm = normalization
        elif callable(normalization):
            norm = [normalization(value) for value in self.x_values]
        else:
            raise ValueError(
                "Invalid normalization:"
                "Options: None, 'max', name of a solution, a number, "
                "an array of numbers, or a callable. See documentation.",
            )
        for _, items in self._solutions.items():
            items.normalize_arf(norm)

    def plot_solutions(
        self,
        ax: None | plt.Axes = None,
        display_values: None | Sequence = None,
        normalization: None | str | float | Callable[[float], float] = None,
        plot_method=plt.plot,
        **kwargs,
    ) -> (plt.Figure, plt.Axes):
        """Plot all solutions in stack over attribute set via
        :meth:`set_abscissa()` or over  ``x_values`` if passed using the
        plotting methods ``plot_method``.
        The plot can be normalized if one of the following is passed
        as ``normalization``:

           - the name of an added solution:
           - ``'max'``: normalization w.r.t. max value of the ARF in the plot
           - a ``float``: normalization w.r.t. to a number
           - a ``callable`` normalization w.r.t. to function that takes the
             values on the x-axis as an input

        :param ax: axes object where plot will be generated
        :param display_values: changing the values for the x-axis
        :param normalization: normalization (see above)
        :param plot_method: matplotlib native plotting method (e.g.plt.loglog)
        :param kwargs: keyword arguments that get piped to :attr:`plot_method`
        """

        # Compute Values
        self._compute_arf()
        self._normalize_arf(normalization)

        # Plot
        if display_values is None:
            display_values = self.x_values
        for _, data in self._solutions.items():
            fig, ax = self.plotter.plot_solution(
                display_values,
                data,
                ax,
                plot_method,
                **kwargs,
            )

        is_normalized = False if normalization is None else True

        self.plotter.set_labels(ax, self.attr_name, is_normalized)
        self.plotter.add_legend(ax)

        return fig, ax


if __name__ == "__main__":
    pass



from __future__ import annotations

from abc import ABC

from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.solutions.basedoinikov1994.arf_general import BaseGeneralARF
from osaft.solutions.doinikov1994rigid.arf_arbitrary_bl import ARFArbitraryBL


class ARFArbitraryRadius(ARFArbitraryBL, BaseGeneralARF, ABC):
    """Child class for ARF according to Doinikov's theory
    (viscous fluid-rigid sphere; 1994)

    .. note::
        This class implements all attributes and methods to compute the ARF
        for the case of an arbitrary particle radius.
        This class does not actually implement a method for the ARF. If you
        want to compute the ARF use
        :attr:`~osaft.solutions.doinikov1994rigid.ARF`

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param position: Position in the standing wave field [rad]
    :param long_wavelength: using :math:`x \\ll 1`
    :param small_boundary_layer: :math:`x \\ll x_v \\ll 1`
    :param large_boundary_layer: :math`x \\ll 1 \\ll x_v`
    :param fastened_sphere: use theory of fastened sphere
    :param background_streaming: background streaming contribution
    :param N_max: Highest order mode
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float,
        long_wavelength: bool,
        small_boundary_layer: bool,
        large_boundary_layer: bool,
        fastened_sphere: bool,
        background_streaming: bool,
        N_max: int,
    ) -> None:
        """Constructor method"""

        # Call to init method of parent class
        ARFArbitraryBL.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
            long_wavelength=long_wavelength,
            small_boundary_layer=small_boundary_layer,
            large_boundary_layer=large_boundary_layer,
            fastened_sphere=fastened_sphere,
            background_streaming=background_streaming,
            N_max=N_max,
        )
        BaseGeneralARF.__init__(self)


if __name__ == "__main__":
    pass



import unittest

import osaft
from osaft.tests.basetest_arf import HelperStandingARF, HelperTravelingARF
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory

conj = osaft.core.functions.conj


class TestARF(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self._v_boundary_conditions = 1e-3
        self.runs = 5

        self.cls = SolutionFactory().doinikov_2021_viscous_arf()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Helper Methods
    # -------------------------------------------------------------------------

    def V_r(self, n: int) -> complex:
        return self.cls.V_r(n, self.cls.R_0, True, True)

    def V_theta(self, n: int) -> complex:
        return self.cls.V_theta(n, self.cls.R_0, True, True)

    # -------------------------------------------------------------------------
    # ARF
    # -------------------------------------------------------------------------

    def compute_arf(self) -> float:
        first_sum = 4 * self.eta_f / self.R_0**2 * self.cls.C_0[
            3,
            1,
        ] + 2 / 3 * self.R_0 * self.rho_f * self.cls.gamma(1, self.R_0, False)

        second_sum = complex(0)
        for n in self.cls.range_N_max:
            if n == 0:
                term_1 = 0
            else:
                term_1 = (
                    n
                    * conj(self.V_r(n - 1))
                    * (self.V_r(n) + (n + 1) * self.V_theta(n))
                    / ((2 * n - 1) * (2 * n + 1))
                )
            term_2 = (
                (n + 1)
                * conj(self.V_r(n + 1))
                * (self.V_r(n) - n * self.V_theta(n))
                / ((2 * n + 1) * (2 * n + 3))
            )
            second_sum += term_1 + term_2
        return (
            2
            * osaft.pi
            * self.R_0**2
            * (first_sum - self.rho_f * second_sum.real)
        )


class TestARFStanding(TestARF, HelperStandingARF):
    def setUp(self) -> None:
        super().setUp()


class TestARFTravelling(TestARF, HelperTravelingARF):
    def setUp(self) -> None:
        super().setUp()


class TestARFErrorEstimation(TestARF):
    def setUp(self) -> None:
        super().setUp()

    def test_error_estimation(self):
        arf, error = self.cls.compute_arf(True)
        self.assertTrue(abs(arf) > 100 * abs(error))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from osaft.core.warnings import AssumptionWarning, raise_assumption_warning


class TestCustomWarnings(unittest.TestCase):
    def test_AssumptionWarning(self) -> None:
        for _ in np.arange(5):
            val = np.random.randint(0, 100) * np.random.random_sample()

            with self.assertWarns(AssumptionWarning) as _:
                raise_assumption_warning(val >= 0.9 * val)
            with self.assertWarns(AssumptionWarning) as _:
                raise_assumption_warning(val <= 1.1 * val)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from osaft.core.backgroundfields import BackgroundField, WaveType
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.fluids import InviscidFluid
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.solids import ElasticSolid
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_solution import BaseSolution


class BaseHasegawa(BaseSphereFrequencyComposite, BaseSolution):
    """Base class for Hasegawa & Yosioka (1969)

    :param f: Frequency [Hz]
    :param R_0: Radius of the particle [m]
    :param rho_s: Density of the solid scattering particle [kg/m^3]
    :param E_s: Young's modulus of the solid scattering particle [N/m^2]
    :param nu_s: Poisson's ratio of the solid scattering particle [-]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound in the fluid [m/s]
    :param p_0: Pressure amplitude of the incident field [Pa]
    :param position: Position in the standing wave field [rad]
    :param wave_type: Progressive wave [-]
    """

    # Supported wave_type for the class
    supported_wavetypes = [WaveType.TRAVELLING, WaveType.STANDING]

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        E_s: float,
        nu_s: float,
        rho_f: float,
        c_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
    ) -> None:

        """Constructor method"""
        BaseSphereFrequencyComposite.__init__(self, f, R_0)
        BaseSolution.__init__(self, "Hasegawa1969")

        #       Initialize Components
        self.scatterer = ElasticSolid(self.frequency, E_s, nu_s, rho_s)
        self.fluid = InviscidFluid(self.frequency, rho_f, c_f)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        #       Dependent variables
        self._lambda_rho = ActiveVariable(
            self._compute_lambda_rho,
            "ratio of densities",
        )
        self._x_f = ActiveVariable(
            self._compute_x_f,
            "dimensionless fluid wavenumber",
        )
        self._x_s_l = ActiveVariable(
            self._compute_x_s_l,
            "dimensionless particle wavenumber, longitudinal direction",
        )
        self._x_s_t = ActiveVariable(
            self._compute_x_s_t,
            "dimensionless particle wavenumber, transversal direction",
        )

        #       Dependencies
        self._lambda_rho.is_computed_by(
            self.scatterer._rho_s,
            self.fluid._rho_f,
        )
        self._x_f.is_computed_by(self.sphere._R_0, self.fluid._k_f)
        self._x_s_l.is_computed_by(self.sphere._R_0, self.scatterer._k_l)
        self._x_s_t.is_computed_by(self.sphere._R_0, self.scatterer._k_t)

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def x_f(self) -> float:
        """dimensionless wave number in the fluid [-]"""
        return self._x_f.value

    @property
    def x_s_l(self) -> float:
        """dimensionless wave number in the particle,
        longitudinal direction [-]
        """
        return self._x_s_l.value

    @property
    def x_s_t(self) -> float:
        """dimensionless wave number in the particle,
        transversal direction [-]
        """
        return self._x_s_t.value

    @property
    def lambda_rho(self) -> float:
        """ratio of densities [-]"""
        return self._lambda_rho.value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to :attr:`osaft.core.backgroundfields.BackgroundField.p_0`"""
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> WaveType:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: WaveType) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.rho_s`"""
        return self.scatterer.rho_s

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_s = value

    @property
    def E_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.E_s`"""
        return self.scatterer.E_s

    @E_s.setter
    def E_s(self, value: float) -> None:
        self.scatterer.E_s = value

    @property
    def nu_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.nu_s`"""
        return self.scatterer.nu_s

    @nu_s.setter
    def nu_s(self, value: float) -> None:
        self.scatterer.nu_s = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`"""
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`"""
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def k_s_l(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.k_l`"""
        return self.scatterer.k_l

    @property
    def k_s_t(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.k_t`"""
        return self.scatterer.k_t

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.kappa_f`"""
        return self.fluid.kappa_f

    @property
    def k_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.k_f`"""
        return self.fluid.k_f

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Field Attributes
    # -------------------------------------------------------------------------

    def A_in(self, n) -> complex:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.A_in`
        """
        return self.field.A_in(n)

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------
    def _compute_lambda_rho(self) -> float:
        """Compute lambda (ratio of densities)"""
        return self.rho_s / self.rho_f

    def _compute_x_f(self) -> float:
        """Compute x_f"""
        return self.R_0 * self.k_f

    def _compute_x_s_l(self) -> float:
        """Compute x_s_l"""
        return self.R_0 * self.k_s_l

    def _compute_x_s_t(self) -> float:
        """Compute x_s_t"""
        return self.R_0 * self.k_s_t


if __name__ == "__main__":
    pass



from __future__ import annotations

import copy
from abc import ABC, abstractmethod

from osaft.core.backgroundfields import WaveType, WrongWaveTypeError


class BaseSolution(ABC):
    """Base class for all solutions

    :param name: name of the solution
    """

    # supported wave types of the solution
    supported_wavetypes = []

    def __init__(self, name: str):
        self.name = name

    def copy(self) -> BaseSolution:
        """Returns a copy of the object

        :rtype: BaseSolution
        """
        return copy.copy(self)

    @property
    @abstractmethod
    def wave_type(self) -> WaveType:
        """returns the wave type of the solution"""
        pass

    def check_wave_type(self) -> None:
        """Checks if :attr:`wave_type` is in :attr:`supported_wavetypes`

        :raises WrongWaveTypeError: If :attr:`wave_type` is not supported
        """
        if self.wave_type not in self.supported_wavetypes:
            raise WrongWaveTypeError(
                "Solution does not exist for "
                f"wave_type = {self.wave_type} \n",
                f"supported: {self.supported_wavetypes}",
            )

    def __copy__(self) -> BaseSolution:
        """Overriding the default copy dunder

        The reason for that is, that we need to create a new object by calling
        the __init__(self, ...) method because otherwise the links to the
        various PassiveVariable and ActiveVariable would be wrong.

        This is circumvent by creating a dictionary with all the possible
        PassiveVariable objects and then calling the __init__(...) method.

        :rtype: BaseSolution
        """
        passive_variables = self.input_variables()
        init_dict = {}
        for name in passive_variables:
            init_dict[name] = getattr(self, name)

        result = self.__class__(**init_dict)
        result.name = f"copy_{self.name}"

        return result


if __name__ == "__main__":
    pass



from __future__ import annotations

import numpy as np

from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.geometries import Sphere
from osaft.core.helper import InputHandler
from osaft.core.variable import ActiveListVariable
from osaft.solutions.basedoinikov1994.scattering import (
    BaseScatteringDoinikov1994,
)
from osaft.solutions.doinikov1994compressible.coefficientmatrix import (
    CoefficientMatrix,
)

NDArray = np.ndarray


class ScatteringField(CoefficientMatrix, BaseScatteringDoinikov1994):
    """Scattering field of Doinikov (viscous fluid-viscous sphere; 1994)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param c_s: Speed of sound of in the sphere [m/s]
    :param eta_s: shear viscosity of in the sphere  [Pa s]
    :param zeta_s: bulk viscosity of in the sphere [Pa s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param position: Position within the standing wave field [m]
    :param wave_type: Type of wave, traveling or standing
    :param N_max: Highest order mode included in the computation [-]
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        c_s: float,
        eta_s: float,
        zeta_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: WaveType,
        position: float,
        N_max: None | int = 5,
    ) -> None:
        """Constructor method"""

        # init of parent class
        CoefficientMatrix.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            c_s=c_s,
            eta_s=eta_s,
            zeta_s=zeta_s,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
        )
        BaseScatteringDoinikov1994.__init__(self, N_max=N_max)

        # Dependent variables
        self._alpha_n = ActiveListVariable(
            self._compute_alpha_n,
            "fluid scattering coefficient alpha_n",
        )
        self._beta_n = ActiveListVariable(
            self._compute_beta_n,
            "fluid scattering coefficient beta_n",
        )
        self._alpha_hat_n = ActiveListVariable(
            self._compute_alpha_hat_n,
            "sphere scattering coefficient alpha_hat_n",
        )
        self._beta_hat_n = ActiveListVariable(
            self._compute_beta_hat_n,
            "sphere scattering coefficient beta_hat_n",
        )

    def __repr__(self):  # pragma: no cover
        return (
            f"Doinikov1994Compressible.ScatteringFiels(f={self.f}, "
            f"R_0={self.R_0}, "
            f"rho_s={self.rho_s}, c_s={self.c_s}, "
            f"eta_s={self.eta_s}, zeta_s={self.zeta_s}, "
            f"rho_f={self.rho_f}, c_f={self.c_f}, "
            f"eta_f={self.eta_f}, zeta_f={self.zeta_f}, "
            f"p_0={self.p_0}, position={self.position}, {self.wave_type}, "
            f"N_max={self.N_max})"
        )

    # -----------------------------------------------------
    # Scattering coefficients fluid
    # -----------------------------------------------------

    def alpha_n(self, n: int) -> complex:
        """coefficient :math:`\\alpha_n` (3.13) and (3.20)

        :param n: order
        """
        return self._alpha_n.item(n)

    def _compute_alpha_n(self, n: int) -> complex:
        # computation according to (3.13), (3.14) and (3.20)
        if n == 0:
            tmp1 = 4 * (self.eta_f - self.eta_s) * self.x * self.x_hat
            tmp1 *= Bf.besselj(1, self.x_hat)

            tmp2 = 1j * self.omega * self.R_0**2

            # numerator
            num = self.rho_f * self.x_hat
            num *= Bf.besselj(0, self.x) * Bf.besselj(1, self.x_hat)
            num -= (
                self.rho_s
                * self.x
                * (Bf.besselj(1, self.x) * Bf.besselj(0, self.x_hat))
            )
            num *= tmp2
            num -= tmp1 * Bf.besselj(1, self.x)

            # denominator
            den = self.rho_s * self.x
            den *= Bf.hankelh1(1, self.x) * Bf.besselj(0, self.x_hat)
            den -= (
                self.rho_f
                * self.x_hat
                * (Bf.hankelh1(0, self.x) * Bf.besselj(1, self.x_hat))
            )
            den *= tmp2
            den += tmp1 * Bf.hankelh1(1, self.x)

            new_value = num / den
        else:
            new_value = self.det_M_n(n, 0) / self.det_M_n(n)

        return new_value

    def beta_n(self, n: int) -> complex:
        """coefficient :math:`\\beta_n` (3.13) and (3.20)

        :param n: order
        """
        return self._beta_n.item(n)

    def _compute_beta_n(self, n: int) -> complex:
        # computation according to (3.13), (3.15) and (3.21)
        if n == 0:
            return 0
        else:
            return self.det_M_n(n, 1) / self.det_M_n(n)

    # -----------------------------------------------------
    # Scattering coefficients sphere
    # -----------------------------------------------------

    def alpha_hat_n(self, n: int) -> complex:
        """coefficient :math:`\\alpha_hat_n` (3.13) and (3.20)

        :param n: order
        """
        return self._alpha_hat_n.item(n)

    def _compute_alpha_hat_n(self, n: int) -> complex:
        # computation according to (3.13), (3.14) and (3.20)

        if n == 0:
            tmp1 = 4 * (self.eta_f - self.eta_s) * self.x * self.x_hat
            tmp2 = 1j * self.omega * self.R_0**2

            # numerator
            num = Bf.besselj(0, self.x) * Bf.hankelh1(1, self.x)

            num -= Bf.besselj(1, self.x) * Bf.hankelh1(0, self.x)

            num *= tmp2 * self.x * self.rho_f

            # denominator
            den = self.rho_s * self.x
            den *= Bf.hankelh1(1, self.x) * Bf.besselj(0, self.x_hat)

            den -= (
                self.rho_f
                * self.x_hat
                * (Bf.hankelh1(0, self.x) * Bf.besselj(1, self.x_hat))
            )
            den *= tmp2
            den += tmp1 * (Bf.hankelh1(1, self.x) * Bf.besselj(1, self.x_hat))

            new_value = num / den
        else:
            new_value = self.det_M_n(n, 2) / self.det_M_n(n)
        return new_value

    def beta_hat_n(self, n: int) -> complex:
        """coefficient :math:`\\beta_hat_n` (3.13) and (3.20)

        :param n: order
        """
        return self._beta_hat_n.item(n)

    def _compute_beta_hat_n(self, n: int) -> complex:
        # computation according to (3.13), (3.15) and (3.21)
        if n == 0:
            return 0
        else:
            return self.det_M_n(n, 3) / self.det_M_n(n)

    # -----------------------------------------------------
    # Methods
    # -----------------------------------------------------

    def radial_particle_velocity(
        self,
        r: float | NDArray | list[float],
        theta: float | NDArray | list[float],
        t: float | NDArray | list[float],
        mode: None | int = None,
    ) -> complex:
        """Particle velocity in radial direction

        Returns the value of the particle velocity
        in radial direction in [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode to be plotted, if `None` then sum of all mode up
                     to `N_max`
        """
        r, theta, t = InputHandler.handle_input(
            r,
            theta,
            t,
            self.R_0,
            inside_sphere=True,
        )

        def radial_velocity(n: int, r: float) -> complex:
            out = self.k_s * self.alpha_hat_n(n)
            out *= Bf.d1_besselj(n, r * self.k_s)

            arg = r * self.k_vs
            out -= n * (n + 1) / r * self.beta_hat_n(n) * Bf.besselj(n, arg)
            out *= self.A_in(n)
            return out

        return self.radial_mode_superposition(
            radial_velocity,
            r,
            theta,
            t,
            mode,
        )

    def tangential_particle_velocity(
        self,
        r: float | NDArray | list[float],
        theta: float | NDArray | list[float],
        t: float | NDArray | list[float],
        mode: None | int = None,
    ) -> complex:
        """Particle velocity in tangential direction

        Returns the value of the particle velocity
        in tangential direction in [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode to be plotted, if `None` then sum of all mode up
                     to `N_max`
        """
        r, theta, t = InputHandler.handle_input(
            r,
            theta,
            t,
            self.R_0,
            inside_sphere=True,
        )

        def tangential_velocity(n: int, r: float) -> complex:
            out = Bf.besselj(n, self.k_vs * r)
            out += r * self.k_vs * Bf.d1_besselj(n, self.k_vs * r)
            out *= -self.beta_hat_n(n)
            out += self.alpha_hat_n(n) * Bf.besselj(n, self.k_s * r)
            out *= self.A_in(n) / r
            return out

        return self.tangential_mode_superposition(
            tangential_velocity,
            r,
            theta,
            t,
            mode,
        )


if __name__ == "__main__":
    pass



import unittest

from osaft.core.frequency import Frequency
from osaft.core.functions import sqrt
from osaft.core.solids import ElasticSolid, RigidSolid
from osaft.tests.basetest_numeric import BaseTestSolutions


class TestConversion(unittest.TestCase):
    def setUp(self) -> None:
        self.rho = 8960
        self.c_1 = 4521.3
        self.c_2 = 2226.1

        self.E = 130e9
        self.nu = 0.34

        self.lam = 1.03e11
        self.G = 4.85e10

    def test_elastic_to_lame(self):
        lam = ElasticSolid.lambda_from_E_nu(self.E, self.nu)
        G = ElasticSolid.G_from_E_nu(self.E, self.nu)
        E_test = ElasticSolid.E_from_Lame(lam, G)
        nu_test = ElasticSolid.nu_from_Lame(lam, G)
        self.assertAlmostEqual(self.E, E_test)
        self.assertAlmostEqual(self.nu, nu_test)

    def test_lame_to_elastic(self):
        E = ElasticSolid.E_from_Lame(self.lam, self.G)
        nu = ElasticSolid.nu_from_Lame(self.lam, self.G)
        lam_test = ElasticSolid.lambda_from_E_nu(E, nu)
        G_test = ElasticSolid.G_from_E_nu(E, nu)
        self.assertAlmostEqual(self.lam, lam_test, delta=1)
        self.assertAlmostEqual(self.G, G_test, delta=1)

    def test_wave_speed_to_elastic(self):
        E = ElasticSolid.E_from_wave_speed(self.c_1, self.c_2, self.rho)
        nu = ElasticSolid.nu_from_wave_speed(self.c_1, self.c_2)
        c_1_test = ElasticSolid.c1_from_E_nu(E, nu, self.rho)
        c_2_test = ElasticSolid.c2_from_E_nu(E, nu, self.rho)
        self.assertAlmostEqual(self.c_1, c_1_test)
        self.assertAlmostEqual(self.c_2, c_2_test)

    def test_elastic_to_wave_speed(self):
        c_1 = ElasticSolid.c1_from_E_nu(self.E, self.nu, self.rho)
        c_2 = ElasticSolid.c2_from_E_nu(self.E, self.nu, self.rho)
        E_test = ElasticSolid.E_from_wave_speed(c_1, c_2, self.rho)
        nu_test = ElasticSolid.nu_from_wave_speed(c_1, c_2)
        self.assertAlmostEqual(self.E, E_test)
        self.assertAlmostEqual(self.nu, nu_test)

    def test_wave_speed_to_lame(self):
        lam = ElasticSolid.lambda_from_wave_speed(self.c_1, self.c_2, self.rho)
        G = ElasticSolid.G_from_wave_speed(self.c_2, self.rho)
        c_1_test = ElasticSolid.c1_from_Lame(lam, G, self.rho)
        c_2_test = ElasticSolid.c2_from_Lame(G, self.rho)
        self.assertAlmostEqual(self.c_1, c_1_test)
        self.assertAlmostEqual(self.c_2, c_2_test)

    def test_lame_to_wave_speed(self):
        c_1 = ElasticSolid.c1_from_Lame(self.lam, self.G, self.rho)
        c_2 = ElasticSolid.c2_from_Lame(self.G, self.rho)
        lam_test = ElasticSolid.lambda_from_wave_speed(c_1, c_2, self.rho)
        G_test = ElasticSolid.G_from_wave_speed(c_2, self.rho)
        self.assertAlmostEqual(self.lam, lam_test, delta=1)
        self.assertAlmostEqual(self.G, G_test, delta=1)


class TestRigidSolid(BaseTestSolutions):
    def setUp(self) -> None:

        super().setUp()

        self.cls = RigidSolid(self.f, self.rho_s)
        self.composite_cls = RigidSolid(self.frequency, self.rho_s)
        self.list_cls = [self.cls, self.composite_cls]

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_properties(self):
        self.do_testing(lambda: self.rho_s, lambda: self.cls.rho_s)
        self.do_testing(
            lambda: self.cls.rho_s,
            lambda: self.composite_cls.rho_s,
        )

        self.do_testing(
            lambda: self.f,
            lambda: self.cls.f,
        )
        self.do_testing(
            lambda: self.cls.f,
            lambda: self.composite_cls.f,
        )


class TestElasticSolid(BaseTestSolutions):
    def setUp(self) -> None:

        super().setUp()

        self.frequency = Frequency(self.f)
        self.cls = ElasticSolid(self.f, self.E_s, self.nu_s, self.rho_s)
        self.composite_cls = ElasticSolid(
            self.frequency,
            self.E_s,
            self.nu_s,
            self.rho_s,
        )
        self.list_cls = [self.cls, self.composite_cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def compute_G(self) -> float:
        return ElasticSolid.G_from_E_nu(self.E_s, self.nu_s)

    def compute_lambda(self) -> float:
        return ElasticSolid.lambda_from_E_nu(self.E_s, self.nu_s)

    def compute_B_s(self) -> float:
        return self.E_s / 3 / (1 - 2 * self.nu_s)

    def compute_kappa_s(self) -> float:
        return 3 * (1 - 2 * self.nu_s) / self.E_s

    def compute_c_l(self) -> float:
        return sqrt(
            self.E_s
            * (1 - self.nu_s)
            / (self.rho_s * (1 + self.nu_s) * (1 - 2 * self.nu_s)),
        )

    def compute_c_t(self) -> float:
        return sqrt(self.E_s / (2 * (1 + self.nu_s) * self.rho_s))

    def compute_k_l(self) -> float:
        omega = self.frequency.omega
        c_l = self.compute_c_l()
        return omega / c_l

    def compute_k_t(self) -> float:
        omega = self.frequency.omega
        c_t = self.compute_c_t()
        return omega / c_t

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_G(self) -> None:
        self.do_testing(
            func_1=self.compute_G,
            func_2=lambda: self.cls.G,
        )

    def test_lambe_2(self) -> None:
        self.do_testing(
            func_1=self.compute_G,
            func_2=lambda: self.cls.lame_2,
        )

    def test_lambda(self) -> None:
        self.do_testing(
            func_1=self.compute_lambda,
            func_2=lambda: self.cls.lame_1,
        )

    def test_c_l(self) -> None:
        self.do_testing(
            func_1=self.compute_c_l,
            func_2=lambda: self.cls.c_l,
        )
        self.do_testing(
            func_1=self.compute_c_l,
            func_2=lambda: self.composite_cls.c_l,
        )

    def test_c_t(self) -> None:
        self.do_testing(
            func_1=self.compute_c_t,
            func_2=lambda: self.cls.c_t,
        )
        self.do_testing(
            func_1=self.compute_c_t,
            func_2=lambda: self.composite_cls.c_t,
        )

    def test_k_l(self) -> None:
        self.do_testing(
            func_1=self.compute_k_l,
            func_2=lambda: self.cls.k_l,
        )
        self.do_testing(
            func_1=self.compute_k_l,
            func_2=lambda: self.composite_cls.k_l,
        )

    def test_k_t(self) -> None:
        self.do_testing(
            func_1=self.compute_k_t,
            func_2=lambda: self.cls.k_t,
        )
        self.do_testing(
            func_1=self.compute_k_t,
            func_2=lambda: self.composite_cls.k_t,
        )

    def test_kappa_s(self) -> None:
        self.do_testing(
            func_1=self.compute_kappa_s,
            func_2=lambda: self.cls.kappa_s,
        )
        self.do_testing(
            func_1=self.compute_kappa_s,
            func_2=lambda: self.composite_cls.kappa_s,
        )

    def test_B_s(self) -> None:
        self.do_testing(
            func_1=self.compute_B_s,
            func_2=lambda: self.cls.B_s,
        )
        self.do_testing(
            func_1=self.compute_B_s,
            func_2=lambda: self.composite_cls.B_s,
        )

    def test_nu_error(self):
        with self.assertRaises(ValueError):
            self.cls.nu_s = -0.1
        with self.assertRaises(ValueError):
            self.cls.nu_s = 0.6


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import cos, exp, sin
from osaft.tests.basetest_scattering import HelperScattering
from osaft.tests.solution_factory import SolutionFactory
from osaft.tests.solutions.king1934.test_base import BaseKing


class TestScattering(HelperScattering, BaseKing):
    def setUp(self) -> None:

        super().setUp()

        self.small_particle_limit = False
        self.cls = SolutionFactory().king_1934_scattering()
        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def A_dash_n(self, n: int):
        arg = self.cls.alpha
        out = -self.cls.A_in(n) * self.G_n(n, arg)
        out /= self.G_n(n, arg) + 1j * self.F_n(n, arg)
        return out

    def potential_coefficient(self, n: int) -> complex:
        return -self.A_dash_n(n)

    def V_r_sc(self, n, r) -> complex:
        out = self.A_dash_n(n)
        out *= self.cls.k_f
        out *= Bf.d1_hankelh2(n, self.cls.k_f * r)
        return out

    def V_theta_sc(self, n, r) -> complex:
        out = self.A_dash_n(n)
        out *= Bf.hankelh2(n, self.cls.k_f * r)
        out /= r
        return out

    def Phi_incident(
        self,
        r,
        theta,
        t,
    ) -> complex:
        def Phi_n(n):
            arg = self.cls.A_in(n) * Bf.besselj(n, r * self.cls.k_f)
            out = exp(-1j * self.cls.omega * t)
            out *= Leg.cos_monomial(n, theta, arg)
            return out

        return np.sum([Phi_n(n) for n in range(self.N_max + 1)])

    def Phi_scattering(
        self,
        r,
        theta,
        t,
    ) -> complex:
        def Phi_n(n):
            arg = self.G_n(n, self.cls.alpha)
            arg += 1j * self.F_n(n, self.cls.alpha)

            arg = -self.G_n(n, self.cls.alpha) / arg

            arg *= self.cls.A_in(n) * Bf.hankelh2(n, r * self.cls.k_f)
            out = exp(-1j * self.cls.omega * t)
            out *= Leg.cos_monomial(n, theta, arg)
            return out

        return np.sum([Phi_n(n) for n in range(self.N_max + 1)])

    def radial_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return self.particle_velocity(t) * cos(theta)
        else:
            return 0

    def tangential_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return -self.particle_velocity(t) * sin(theta)
        else:
            return 0

    def particle_velocity(self, t):
        alpha = self.cls.alpha

        out = self.F_n(1, alpha)
        out -= 1j * self.G_n(1, alpha)
        out = 1 / out

        out *= self.cls.A_in(1) / alpha**3
        out *= self.cls.k_f * self.rho_f / self.rho_s
        out *= exp(-1j * self.cls.omega * t)

        return out

    # -------------------------------------------------------------------------
    # Test coefficients
    # -------------------------------------------------------------------------

    def test_methods(self) -> None:
        method_dict = {}
        method_dict["A_dash_n"] = None
        method_dict["potential_coefficient"] = None

        # needs to be a float otherwise
        # Integers to negative integer powers are not allowed.
        method_dict["phi_n"] = 2.0
        method_dict["psi_n"] = 2.0

        arg = self.cls.k_f * self.get_random_r(self.cls)
        method_dict["F_n"] = arg
        method_dict["G_n"] = arg

        arg = self.get_random_r(self.cls)
        method_dict["V_r_sc"] = arg
        method_dict["V_theta_sc"] = arg

        self._test_methods_n(method_dict)

    def test_Phi_incident(self) -> None:
        for i in range(5):
            r = self.get_random_r(self.cls)
            t = self.get_random_t()
            theta = self.get_random_theta()
            self.parameters.N_max = i
            self.cls.N_max = i

            self.do_testing(
                func_1=self.Phi_incident,
                args_1=(r, theta, t),
                func_2=self.cls.Phi_incident,
                args_2=(r, theta, t),
            )

    def test_Phi_scattering(self) -> None:
        for i in range(5):
            r = self.get_random_r(self.cls)
            t = self.get_random_t()
            theta = self.get_random_theta()

            self.parameters.N_max = i
            self.cls.N_max = i

            self.do_testing(
                func_1=self.Phi_scattering,
                args_1=(r, theta, t),
                func_2=self.cls.Phi_scattering,
                args_2=(r, theta, t),
            )


if __name__ == "__main__":
    unittest.main()



import unittest

from osaft.tests.solution_factory import SolutionFactory
from osaft.tests.solutions.basedoinikov1994.test_base import (
    TestBaseDoinikov1994,
)


class BaseTestARF(TestBaseDoinikov1994):
    def setUp(self) -> None:
        super().setUp()

        self.long_wavelength = False
        self.small_boundary_layer = False
        self.large_boundary_layer = False
        self.background_streaming = True
        self.fastened_sphere = False

        self.parameters._eta_f.low = 1e-5
        self.parameters._eta_f.high = 1e-3
        self.parameters._f.low = 1e6
        self.parameters._f.high = 10e6
        self.parameters._R_0.low = 1e-6

        self.cls = SolutionFactory().doinikov_1994_rigid_arf()
        self.cls.long_wavelength = self.long_wavelength
        self.cls.small_boundary_layer = self.small_boundary_layer
        self.large_boundary_layer = self.large_boundary_layer
        self.fastened_sphere = self.fastened_sphere

        self.list_cls = [self.cls]

    def assign_parameters(self) -> None:
        super().assign_parameters()

        self.cls.fastened_sphere = self.fastened_sphere
        self.cls.small_boundary_layer = self.small_boundary_layer
        self.cls.large_boundary_layer = self.large_boundary_layer
        self.cls.long_wavelength = self.long_wavelength
        self.cls.background_streaming = self.background_streaming
        self.cls.N_max = self.N_max


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

import numpy as np

from osaft import log
from osaft.core.functions import exp, pi
from osaft.core.functions import (
    spherical_2_cartesian_coordinates as s2c_coord,
)
from osaft.core.functions import spherical_2_cartesian_vector as s2c_vector
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.base_scattering import BaseScattering

NDArray = np.ndarray
# -----------------------------------------------------------------------------
# ParticleWireFrameData
# -----------------------------------------------------------------------------


@dataclass
class DeformedLine:
    """Class for storing and retrieving the reference configuration and the
    deformed configuration of a deformed line

    :param x: x-coordinate of the reference configuration
    :param y: y-coordinate of the reference configuration
    :param u: displacement in x-direction
    :param y: displacement in y-direction
    """

    def __init__(
        self,
        x: NDArray,
        y: NDArray,
        u: NDArray,
        v: NDArray,
    ) -> None:
        """Constructor method"""
        self.x = x
        self.y = y
        self.u = u
        self.v = v

    def get_deformed(self, phase) -> tuple[NDArray, NDArray]:
        """Returns the deformed line"""
        x_deformed = self.x + (self.u * exp(1j * phase)).real
        y_deformed = self.y + (self.v * exp(1j * phase)).real
        return x_deformed, y_deformed

    def get_reference(self) -> tuple[NDArray, NDArray]:
        return self.x, self.y


class ParticleWireframeData:
    """Data container for wireframe plot of the particle

    :param sol: solution to be plotted
    :param nbr_r_levels: number of circles shown in the wireframe
    :param nbr_theta_levels: number of radii shown in the wireframe
    :param resolution: resolution, if tuple `(radial res, tangential res)`
    :param scale_factor: ratio between shown displacement and radius
    :param symmetric: if `True` plot is expected to be symmetric.
    """

    def __init__(
        self,
        sol: BaseScattering,
        nbr_r_levels: int,
        nbr_theta_levels: int,
        resolution: int | tuple[int, int] = (100, 100),
        scale_factor: float = 0.1,
        symmetric: bool = True,
    ) -> None:
        """Constructor method"""
        # Solution
        self._sol = PassiveVariable(sol, "Solution for plotting")
        self._mode = PassiveVariable(None, "mode")

        # Resolution
        unpacked_resolution = self._unpack_resolution(resolution)
        self._resolution = PassiveVariable(unpacked_resolution)

        # Scaling factor
        self._rel_scale_factor = PassiveVariable(
            scale_factor,
            "Relative scale factor",
        )
        self._scale_factor = ActiveVariable(
            self._compute_scale_factor,
            "Absolute scale factor",
        )
        self._scale_factor.is_computed_by(
            self._sol,
            self._rel_scale_factor,
            self._mode,
        )

        # Symmetry
        self._symmetric = PassiveVariable(symmetric, "symmetric")

        # Coordinate Grid
        self._arr_r = ActiveVariable(
            self._compute_arr_r,
            "Coordinate grid in radial direction",
        )
        self._arr_theta = ActiveVariable(
            self._compute_arr_theta,
            "Coordinate grid in radial direction",
        )
        self._arr_r.is_computed_by(self._resolution, self._sol)
        self._arr_theta.is_computed_by(
            self._resolution,
            self._symmetric,
            self._sol,
        )

        # Number of Levels
        self._nbr_r_lvl = PassiveVariable(
            nbr_r_levels,
            "Number of radial wireframe " "levels",
        )
        even_nbr_theta_levels = self._test_nbr_t_lvl(nbr_theta_levels)
        self._nbr_t_lvl = PassiveVariable(
            even_nbr_theta_levels,
            "Number of tangential wireframe " "levels",
        )

        # Levels
        self._r_levels = ActiveVariable(
            self._compute_r_levels,
            "Radii of wireframe circles",
        )
        self._theta_levels = ActiveVariable(
            self._compute_theta_levels,
            "Angles of wireframe radii",
        )
        self._r_levels.is_computed_by(self._sol, self._nbr_r_lvl)
        self._theta_levels.is_computed_by(
            self._sol,
            self._nbr_t_lvl,
            self._symmetric,
        )

        # Solution containers
        self._deformed_radii = ActiveVariable(
            self._reset_list,
            "List of deformed radii",
        )
        self._deformed_circles = ActiveVariable(
            self._reset_list,
            "List of deformed circles",
        )
        self._deformed_radii.is_computed_by(
            self._sol,
            self._scale_factor,
            self._arr_r,
            self._theta_levels,
        )
        self._deformed_circles.is_computed_by(
            self._sol,
            self._scale_factor,
            self._arr_theta,
            self._r_levels,
        )

    # -------------------------------------------------------------------------
    # __init__ helper methods
    # -------------------------------------------------------------------------

    def _test_nbr_t_lvl(self, nbr_t_lvl: int) -> int:
        nbr_t_lvl = int(nbr_t_lvl)
        if nbr_t_lvl % 2 and self.symmetric:
            log.warning(
                "nbr_theta_levels has to be even. nbr_theta_levels is set "
                f"to {nbr_t_lvl} + 1 = {nbr_t_lvl + 1} ",
            )
            return nbr_t_lvl + 1
        else:
            return int(nbr_t_lvl)

    @staticmethod
    def _unpack_resolution(
        res: int | tuple[int, int],
    ) -> tuple[int, int]:
        """Unpacks resolution tuple if needed

        if `res` is an `int` a tuple `(res, res)` is
        returned. If `res` is a tuple with two values, `res` is passed through.
        :param res: res of the grid
        """
        if isinstance(res, Sequence):
            if len(res) == 1:
                return res[0], res[0]
            elif len(res) == 2:
                return res
            else:
                raise ValueError(
                    "Resolution needs to be either one value for both "
                    "radial and tangential direction",
                )
        else:
            return res, res

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_displacements(self) -> tuple[list, list]:
        """Returns the wireframe

        The wireframe is return as a tuple of two list.
        `(deformed_radii, deformed_circle)`. Each element of the list is of
        type :class:`osaft.plotting.wireframe_datacontainer.DeformedLine`.
        """
        return self.deformed_radii, self.deformed_circles

    # -------------------------------------------------------------------------
    # Getters and Setters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def sol(self) -> BaseScattering:
        """Solution for plotting

        :getter: returns the solution for plotting
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._sol.value

    @sol.setter
    def sol(self, value: BaseScattering) -> None:
        self._sol.value = value

    @property
    def resolution(self) -> tuple[int, int]:
        """Plotting resolution

        The plotting resolution of the form
        `(radial_resolution, tangential_resolution)`

        :getter: returns the plotting resolution
        :setter: unpacks resolution and automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._resolution.value

    @resolution.setter
    def resolution(self, value: int | tuple[int, int]) -> None:
        unpacked_value = self._unpack_resolution(value)
        self._resolution.value = unpacked_value

    @property
    def rel_scale_factor(self) -> float:
        """Relative scale factor

        Ratio between max shown displacement and particle radius.
        :getter: returns the relative scale factor
        :setter: unpacks resolution and automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._rel_scale_factor.value

    @rel_scale_factor.setter
    def rel_scale_factor(self, value: float) -> None:
        self._rel_scale_factor.value = value

    @property
    def mode(self) -> int:
        """Mode that is plotted. If `None` all modes up to `sol.N_max`
        are superimposed.

        :getter: returns the mode
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._mode.value

    @mode.setter
    def mode(self, value: int) -> None:
        self._mode.value = value

    @property
    def nbr_r_levels(self) -> int:
        """Number of radial levels in the wireframe plot

        :getter: returns the number of radial plotting levels
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._nbr_r_lvl.value

    @nbr_r_levels.setter
    def nbr_r_levels(self, value: int) -> None:
        self._nbr_r_lvl.value = value

    @property
    def nbr_theta_levels(self) -> int:
        """Number of tangential levels in the wireframe plot

        :getter: returns the number of tangential plotting levels
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._nbr_t_lvl.value

    @nbr_theta_levels.setter
    def nbr_theta_levels(self, value: int) -> None:
        even_number_theta_levels = self._test_nbr_t_lvl(value)
        self._nbr_t_lvl.value = even_number_theta_levels

    @property
    def symmetric(self):
        """Option if the plot is symmetric

        If `True` only have of the data is stored.

        :getter: returns the symmetry bool
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._symmetric.value

    @symmetric.setter
    def symmetric(self, value: bool):
        self._symmetric.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def scale_factor(self) -> float:
        """Returns the scale factor for the deformation of the wire frame
        plot"""
        return self._scale_factor.value

    @property
    def r_levels(self) -> NDArray:
        """Radial levels of the wireframe plot"""
        return self._r_levels.value

    @property
    def theta_levels(self) -> NDArray:
        """Radial levels of the wireframe plot"""
        return self._theta_levels.value

    @property
    def deformed_radii(self) -> list[DeformedLine]:
        """List of deformed radii of the wireframe plot"""
        if not self._deformed_radii.value:
            self._compute_data()
        return self._deformed_radii.value

    @property
    def deformed_circles(self) -> list[DeformedLine]:
        """List of deformed circles of the wireframe plot"""
        if not self._deformed_circles.value:
            self._compute_data()
        return self._deformed_circles.value

    @property
    def arr_r(self) -> NDArray:
        """Radial coordinate grid"""
        return self._arr_r.value

    @property
    def arr_theta(self) -> NDArray:
        """Tangential coordinate grid"""
        return self._arr_theta.value

    # -------------------------------------------------------------------------
    # Methods for dependent variables
    # -------------------------------------------------------------------------

    @staticmethod
    def _reset_list() -> list:
        """Resets ActiveVariable to an empty list"""
        return []

    def _compute_scale_factor(
        self,
    ) -> float:
        """Computes the scaling factor for the wireframe plot

        Compute an appropriate scaling factor for the wireframe plot.
        """
        max_r_displacement = np.nanmax(
            np.abs(
                self.sol.radial_particle_displacement(
                    self.arr_r,
                    self.arr_theta,
                    t=0,
                    mode=self.mode,
                ),
            ),
        )
        max_theta_displacement = np.nanmax(
            np.abs(
                self.sol.tangential_particle_displacement(
                    self.arr_r,
                    self.arr_theta,
                    t=0,
                    mode=self.mode,
                ),
            ),
        )
        max_disp = max(max_r_displacement, max_theta_displacement)
        scale_factor = self.rel_scale_factor * self.sol.R_0 / max_disp
        return scale_factor

    def _compute_r_levels(self) -> NDArray:
        """Compute radial levels"""
        dr = self.sol.R_0 / self.nbr_r_levels
        return np.linspace(dr, self.sol.R_0, self.nbr_r_levels)

    def _compute_theta_levels(self) -> NDArray:
        """Compute tangential levels"""
        if self.symmetric:
            return np.linspace(0, pi, int(self.nbr_theta_levels / 2))
        else:
            return np.linspace(0, 2 * pi, int(self.nbr_theta_levels))

    def _compute_arr_r(self) -> NDArray:
        return np.linspace(1e-30, self.sol.R_0, self.resolution[0])

    def _compute_arr_theta(self) -> NDArray:
        if self.symmetric:
            return np.linspace(0, pi, self.resolution[1])
        else:
            return np.linspace(0, 2 * pi, self.resolution[1])

    # -------------------------------------------------------------------------
    # Compute data
    # -------------------------------------------------------------------------

    def _compute_data(self) -> None:
        self._compute_deformed_radii()
        self._compute_deformed_circles()

    def _compute_deformed_radii(self) -> None:
        for theta in self.theta_levels:
            deformed = self._compute_deformed_radius(theta)
            self._deformed_radii.value.append(deformed)

    def _compute_deformed_circles(self) -> None:
        for r in self.r_levels:
            self._deformed_circles.value.append(
                self._compute_deformed_circle(r),
            )

    def _compute_deformed_radius(self, Theta) -> DeformedLine:
        # Transform coordinates
        x, y = s2c_coord(self.arr_r, Theta)
        # Displacements
        u_r = self.sol.radial_particle_displacement(
            r=self.arr_r,
            theta=Theta,
            t=0,
            mode=self.mode,
        )
        u_theta = self.sol.tangential_particle_displacement(
            r=self.arr_r,
            theta=Theta,
            t=0,
            mode=self.mode,
        )
        # Transform displacements
        u, v = s2c_vector(u_r, u_theta, Theta)
        u_scaled, v_scaled = u * self.scale_factor, v * self.scale_factor
        return DeformedLine(x, y, u_scaled, v_scaled)

    def _compute_deformed_circle(
        self,
        R: float,
    ) -> DeformedLine:
        """Compute the deformed circle for radius `R`
        :param R: radius of the deformed circle
        """
        # Transform coordinates
        x, y = s2c_coord(R, self.arr_theta)
        u_r = self.sol.radial_particle_displacement(
            r=R,
            theta=self.arr_theta,
            t=0,
            mode=self.mode,
        )
        u_theta = self.sol.tangential_particle_displacement(
            r=R,
            theta=self.arr_theta,
            t=0,
            mode=self.mode,
        )
        # Transform displacements
        u, v = s2c_vector(u_r, u_theta, self.arr_theta)
        u_scaled, v_scaled = u * self.scale_factor, v * self.scale_factor
        return DeformedLine(x, y, u_scaled, v_scaled)


if __name__ == "__main__":
    pass



from osaft import WaveType
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestCompareSettnesStanding(BaseTestSolutions):
    def setUp(self):
        super().setUp()

        self.parameters.wave_type = WaveType.STANDING

        # Viscosity
        self.arf_compare_threshold = 10e-1
        self.small_boundary_layer = False
        self.large_boundary_layer = False

        # Viscosity
        self.parameters.eta_s = 1e-3
        self.parameters.zeta_s = 0
        self.parameters.zeta_f = 0
        self.parameters.eta_f = 1e-4
        self.parameters.R_0 = 5e-6

        # Frequency
        self.parameters.f = 2e6

        # Density
        self.parameters.rho_s = 2500

        self.cls = SolutionFactory().doinikov_1994_compressible_arf()

        self.compare_cls = SolutionFactory().settnes_2012_arf()

        self.list_cls = [self.cls, self.compare_cls]

    def test_comparison_arf(self):
        self.assertAlmostEqual(self.cls.compute_arf(), self.cls.compute_arf())


if __name__ == "__main__":
    pass



from __future__ import annotations

from abc import ABC, abstractmethod

from osaft import (
    BackgroundField,
    ElasticSolid,
    ViscoelasticFluid,
    ViscousFluid,
    WaveType,
    log,
)
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_solution import BaseSolution


class BaseDoinikov(BaseSphereFrequencyComposite, BaseSolution, ABC):
    """Base class for Doinikov (viscous fluid-elastic sphere; 2021)

    :param f: frequency [Hz]
    :param R_0: radius [m]
    :param rho_s: density of the particle [kg/m^3]
    :param E_s: Young's modulus of the particle [Pa]
    :param nu_s: Poisson's ratio of the particle [-]
    :param fluid: Fluid instance of the model
    :param p_0: pressure amplitude [Pa]
    :param wave_type: wave type
    :param position: position in the standing wave field [m]
    """

    # Supported wave_type for the class
    supported_wavetypes = [WaveType.STANDING, WaveType.TRAVELLING]

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float,
        rho_s: float,
        E_s: float,
        nu_s: float,
        fluid: ViscousFluid | ViscoelasticFluid,
        p_0: float,
        wave_type: WaveType,
        position: None | float,
    ) -> None:
        """Constructor method"""

        # Init parent classes
        BaseSphereFrequencyComposite.__init__(self, f, R_0)
        BaseSolution.__init__(self, "Doinikov2021Viscous")

        # Initializing fluid
        self.fluid = fluid
        # Making sure fluid has the same Frequency instance
        self.fluid.frequency = self.frequency

        # Initialize Components
        self.solid = ElasticSolid(self.frequency, E_s, nu_s, rho_s)
        self.field = BackgroundField(
            self.fluid,
            p_0,
            wave_type,
            position,
        )

        # Dependent variables
        self._x_l = ActiveVariable(
            self._compute_x_l,
            "dimensionless wavenumber x_l",
        )
        self._x_t = ActiveVariable(
            self._compute_x_t,
            "dimensionless wavenumber x_t",
        )

        self._x_f = ActiveVariable(
            self._compute_x_f,
            "dimensionless wavenumber x_f",
        )
        self._x_v = ActiveVariable(
            self._compute_x_v,
            "dimensionless wavenumber x_v",
        )

        # Dependencies
        self._x_f.is_computed_by(self.fluid._k_f, self.sphere._R_0)
        self._x_v.is_computed_by(self.fluid._k_v, self.sphere._R_0)
        self._x_l.is_computed_by(self.solid._k_l, self.sphere._R_0)
        self._x_t.is_computed_by(self.solid._k_t, self.sphere._R_0)

        log.debug(repr(self))
        log.info(str(self))

    def __repr__(self):  # pragma: no cover
        return (
            f"BaseDoinikov({self.f}, {self.R_0}, {self.rho_s}, "
            f"{self.E_s}), {self.nu_s}. {self.rho_f}, {self.c_f}, "
            f"{self.eta_f}, {self.zeta_f}"
            f", {self.p_0}, {self.wave_type}, "
            f"{self.field.position})"
        )

    def __str__(self):
        out = "Doinikov solution with following properties: \n"
        out += SF.get_str_text("Frequency", "f", self.wave_type, "Hz")
        out += SF.get_str_text(
            "Wavelength",
            "lambda_f",
            self.fluid.lambda_f,
            "m",
        )
        out += SF.get_str_text(
            "Boundary Layer Thickness",
            "delta",
            self.fluid.delta,
            "m",
        )
        out += SF.get_str_text(
            "Viscous wavelength",
            "lambda_v",
            self.lambda_v,
            "m",
        )
        out += SF.get_str_text("Radius", "R_0", self.R_0, "m")
        out += SF.get_str_text(
            "Fluid shear viscosity",
            "eta_f",
            self.eta_f,
            "Pa s",
        )
        out += SF.get_str_text(
            "Fluid bulk viscosity",
            "zeta_f",
            self.zeta_f,
            "Pa s",
        )
        return out

    # -------------------------------------------------------------------------
    # Abstract Methods
    # -------------------------------------------------------------------------

    @property
    @abstractmethod
    def eta(self) -> float | complex:
        """Fluid shear viscosity. For a
        :class:`osaft.core.fluids.ViscousFluid` returns :math:`\\eta_f`, for a
        :class:`osaft.core.fluids.ViscoelasticFluid` returns :math:`\\eta_c`.
        """
        pass

    @property
    @abstractmethod
    def zeta(self) -> float | complex:
        """Fluid bulk viscosity. For a
        :class:`osaft.core.fluids.ViscousFluid` returns :math:`\\zeta_f`, for a
        :class:`osaft.core.fluids.ViscoelasticFluid` returns :math:`\\zeta_c`.
        """
        pass

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def x_l(self):
        """Dimensionless primary wavenumber in the solid

        :math:`k_l \\cdot R_0`

        """
        return self._x_l.value

    @property
    def x_t(self):
        """Dimensionless secondary wavenumber in the solid

        :math:`k_t \\cdot R_0`

        """
        return self._x_t.value

    @property
    def x_f(self):
        """Dimensionless acoustic wavenumber in the fluid

        :math:`k_f \\cdot R_0`

        """
        return self._x_f.value

    @property
    def x_v(self):
        """Dimensionless viscous wavenumber in the fluid

        :math:`k_v \\cdot R_0`

        """
        return self._x_v.value

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_x_l(self) -> complex:
        return self.R_0 * self.k_l

    def _compute_x_t(self) -> complex:
        return self.R_0 * self.k_t

    def _compute_x_f(self) -> complex:
        return self.R_0 * self.k_f

    def _compute_x_v(self) -> complex:
        return self.R_0 * self.k_v

    # -------------------------------------------------------------------------
    # Wrappers for Independent Solid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.rho_s`"""
        return self.solid.rho_s

    @rho_s.setter
    def rho_s(self, value) -> None:
        self.solid.rho_s = value

    @property
    def E_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.E_s`"""
        return self.solid.E_s

    @E_s.setter
    def E_s(self, value) -> None:
        self.solid.E_s = value

    @property
    def nu_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.nu_s`"""
        return self.solid.nu_s

    @nu_s.setter
    def nu_s(self, value) -> None:
        self.solid.nu_s = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Solid Attributes
    # -------------------------------------------------------------------------

    @property
    def k_l(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.k_l`"""
        return self.solid.k_l

    @property
    def k_t(self) -> float:
        """Wraps to :attr:`osaft.core.solids.ElasticSolid.k_t`"""
        return self.solid.k_t

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.rho_0` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.rho_0`
        """
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.c_f` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.c_f`
        """
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value) -> None:
        self.fluid.c_f = value

    @property
    def eta_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.eta_f` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.eta_f`
        """
        return self.fluid.eta_f

    @eta_f.setter
    def eta_f(self, value) -> None:
        self.fluid.eta_f = value

    @property
    def zeta_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.zeta_f` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.zeta_f`
        """
        return self.fluid.zeta_f

    @zeta_f.setter
    def zeta_f(self, value) -> None:
        self.fluid.zeta_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def k_f(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_f` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.k_f`
        """
        return self.fluid.k_f

    @property
    def k_v(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_v` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.k_v`
        """
        return self.fluid.k_v

    @property
    def delta(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.delta` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.delta`
        """
        return self.fluid.delta

    @property
    def lambda_v(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.lambda_v` or to
        :attr:`osaft.core.fluids.ViscoelasticFluid.lambda_v`
        """
        return self.fluid.lambda_v

    # -------------------------------------------------------------------------
    # Wrappers for Independent Background Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.p_0`
        """
        return self.field.p_0

    @p_0.setter
    def p_0(self, value) -> None:
        self.fluid.p_0 = value

    @property
    def wave_type(self) -> WaveType:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Background Field Methods
    # -------------------------------------------------------------------------

    def A_in(self, n):
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.A_in`
        """
        return self.field.A_in(n)


if __name__ == "__main__":
    pass



import matplotlib.pyplot as plt
from matplotlib.quiver import Quiver

from osaft.plotting.datacontainers.scattering_datacontainer import (
    ScatteringFieldData,
)


class BaseScatteringPlots:
    """Base class for the scattering plots. Defines :meth:`_overlay_quiver`"""

    data: ScatteringFieldData

    def _overlay_quiver(
        self,
        ax: plt.Axes,
        phase: float,
        scattered: bool,
        incident: bool,
        mode: int,
        symmetric: bool,
        quiver_color: str,
        animation: bool,
    ) -> Quiver:
        Z, X, U, W, scale = self.data.get_velocity_quiver(
            phase,
            mode,
            scattered,
            incident,
            symmetric,
            animation,
        )

        Q = ax.quiver(
            Z,
            X,
            U,
            W,
            color=quiver_color,
            pivot="tail",
            scale_units="width",
            scale=scale,
        )

        return Q



from __future__ import annotations

from osaft import BackgroundField, ViscousFluid, WaveType
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveVariable


class BaseDoinikov1994(BaseSphereFrequencyComposite):
    """Base class for Doinikov solutions from 1994

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, traveling or standing
    :param position: Position in the standing wave field [rad]
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: None | WaveType = WaveType.STANDING,
        position: None | float = None,
    ) -> None:

        super().__init__(f, R_0)

        self.fluid = ViscousFluid(self.frequency, rho_f, c_f, eta_f, zeta_f)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        # Dependent Variables
        self._x = ActiveVariable(
            self._compute_x,
            "Dimensionless wavenumber",
        )
        self._x_0 = ActiveVariable(
            self._compute_x_0,
            "Real part of the dimensionless wavenumber ",
        )
        self._x_v = ActiveVariable(
            self._compute_x_v,
            "Dimensionless viscous wavenumber",
        )
        self._norm_delta = ActiveVariable(
            self._compute_norm_delta,
            "normalized boundary layer thickness",
        )
        self._x.is_computed_by(
            self.fluid._k_f,
            self.sphere._R_0,
        )
        self._x_0.is_computed_by(
            self._x,
        )
        self._x_v.is_computed_by(
            self.fluid._k_v,
            self.sphere._R_0,
        )
        self._norm_delta.is_computed_by(
            self.fluid._delta,
            self.sphere._R_0,
        )

    # -------------------------------------------------------------------------
    # Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def x(self) -> complex:
        """Product of :attr:`~.k_f` and :attr:`~.R_0`
        :math:`\\hat{x}=k_f R_0`
        """
        return self._x.value

    def _compute_x(self) -> complex:
        return self.k_f * self.R_0

    @property
    def x_v(self) -> complex:
        """Product of :attr:`~.k_v` and :attr:`~.R_0`
        :math:`x_v=k_v R_0`
        """
        return self._x_v.value

    def _compute_x_v(self) -> complex:
        return self.k_v * self.R_0

    @property
    def x_0(self) -> complex:
        """Real part of :math:`x`"""
        return self._x_0.value

    def _compute_x_0(self) -> complex:
        return self.x.real

    @property
    def norm_delta(self) -> float:
        """normalized viscous boundary thickness
        :math:`\\tilde{\\delta}=\\frac{\\delta}{R_0}`
        """
        return self._norm_delta.value

    def _compute_norm_delta(self):
        return self.delta / self.R_0

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.p_0`
        """
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> WaveType:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: WaveType) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.rho_f`"""
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.c_f`"""
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    @property
    def eta_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.eta_f`"""
        return self.fluid.eta_f

    @eta_f.setter
    def eta_f(self, value: float) -> None:
        self.fluid.eta_f = value

    @property
    def zeta_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.zeta_f`"""
        return self.fluid.zeta_f

    @zeta_f.setter
    def zeta_f(self, value: float) -> None:
        self.fluid.zeta_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_t(self) -> float:
        """Returns the ratio of the densities
        :math:`\\tilde{\\rho}=\\frac{\\rho_f}{\\rho_s}`
        """
        return self._rho_t.value

    def _compute_rho_t(self) -> float:
        return self.rho_f / self.rho_s

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.kappa_f`"""
        return self.fluid.kappa_f

    @property
    def k_f(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_f`"""
        return self.fluid.k_f

    @property
    def k_v(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_v`"""
        return self.fluid.k_v

    @property
    def delta(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.delta`"""
        return self.fluid.delta

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def abs_pos(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.abs_pos`
        """
        return self.field.abs_pos

    def A_in(self, n: int) -> complex:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.A_in`

        :param n: mode number
        """
        return self.field.A_in(n)


if __name__ == "__main__":
    pass



from __future__ import annotations

from osaft import log
from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import exp
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveListVariable, ActiveVariable
from osaft.solutions.base_scattering import BaseScatteringRigidParticle
from osaft.solutions.basedoinikov1994.scattering import (
    BaseScatteringDoinikov1994,
)
from osaft.solutions.doinikov1994rigid.base import BaseDoinikov1994Rigid


class ScatteringField(
    BaseDoinikov1994Rigid,
    BaseScatteringRigidParticle,
    BaseScatteringDoinikov1994,
):
    """Scattering field class for Doinikov (viscous fluid-rigid sphere; 1994)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, traveling or standing
    :param position: Position in the standing wave field [rad]
    :param N_max: Highest order mode included in the computation [-]
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: None | WaveType = WaveType.STANDING,
        position: None | float = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method"""

        # init of parent class
        BaseDoinikov1994Rigid.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
        )
        BaseScatteringDoinikov1994.__init__(self, N_max=N_max)

        if type(self) is ScatteringField:
            log.debug(repr(self))
            log.info(str(self))

        # Dependent variables
        self._alpha_n = ActiveListVariable(
            self._compute_alpha_n,
            "scattering coefficient alpha_n",
        )

        self._beta_n = ActiveListVariable(
            self._compute_beta_n,
            "scattering coefficient beta_n",
        )

        self._gamma_n = ActiveListVariable(self._compute_gamma_n)
        self._xi_n = ActiveListVariable(self._compute_xi_n)

        self._mu_1 = ActiveVariable(self._compute_mu_1)
        self._mu_2 = ActiveVariable(self._compute_mu_2)
        self._mu_3 = ActiveVariable(self._compute_mu_3)
        self._mu_4 = ActiveVariable(self._compute_mu_4)

        # define dependencies
        self._mu_1.is_computed_by(
            self._rho_t,
            self._x,
        )
        self._mu_2.is_computed_by(
            self._rho_t,
            self._x,
        )
        self._mu_3.is_computed_by(
            self._rho_t,
            self._x_v,
        )
        self._mu_4.is_computed_by(
            self._rho_t,
            self._x_v,
            self._x,
            self._mu_2,
            self._mu_3,
        )

        self._gamma_n.is_computed_by(self._x_v)

        self._xi_n.is_computed_by(
            self._x,
            self._x_v,
            self._gamma_n,
        )

        self._alpha_n.is_computed_by(
            self._gamma_n,
            self._xi_n,
            self._x,
            self._x_v,
            self._mu_1,
            self._mu_3,
            self._mu_4,
        )

        self._beta_n.is_computed_by(
            self._gamma_n,
            self._xi_n,
            self._x,
            self._x_v,
            self._mu_1,
            self._mu_2,
            self._mu_4,
        )

    def __repr__(self):
        return (
            f"Donikov1994Rigid.ScatteringFiels(f={self.f}, "
            f"R_0={self.R_0}, "
            f"rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, "
            f"eta_f={self.eta_f}, zeta_f={self.zeta_f}, "
            f"p_0={self.p_0}, position={self.position}, {self.wave_type}, "
            f"N_max={self.N_max})"
        )

    def __str__(self):
        out = "Doinikov's(1994 rigid - viscous) solution "
        out += "to the scattering field"
        out += " with following properties: \n"
        out += SF.get_str_text("Wave type", "", self.wave_type, "")
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "rad",
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_0",
            self.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Viscosity",
            "eta_f",
            self.eta_f,
            "Pa s",
        )
        out += SF.get_str_text(
            "Bulk Viscosity",
            "zeta_f",
            self.zeta_f,
            "Pa s",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        out += "Other\n"
        out += SF.get_str_text(
            "#modes",
            "N_max",
            self.N_max,
            None,
        )
        return out

    # -----------------------------------------------------
    # Dependent Variables
    # -----------------------------------------------------

    def xi_n(self, n: int) -> complex:
        """coefficient :math:`\\xi_n` (3.22)

        :param n: order
        """
        return self._xi_n.item(n)

    def _compute_xi_n(self, n: int) -> complex:
        # computation according to (3.13) and (3.20)
        new_value = -n * (n + 1)
        new_value *= Bf.hankelh1(n, self.x) * Bf.hankelh1(n, self.x_v)
        new_value += self.x * Bf.d1_hankelh1(n, self.x) * self.gamma_n(n)

        return new_value

    def gamma_n(self, n: int) -> complex:
        """coefficient :math:`\\gamma_n` (3.22)

        :param n: order
        """
        return self._gamma_n.item(n)

    def _compute_gamma_n(self, n: int) -> complex:
        # computation according to (3.13) and (3.20)
        new_value = Bf.hankelh1(n, self.x_v)
        new_value += self.x_v * Bf.d1_hankelh1(n, self.x_v)
        return new_value

    def alpha_n(self, n: int) -> complex:
        """coefficient :math:`\\alpha_n` (3.13) and (3.20)

        :param n: order
        """
        return self._alpha_n.item(n)

    def _compute_alpha_n(self, n: int) -> complex:
        # computation according to (3.13), (3.14) and (3.20)
        if n == 0:
            new_value = -Bf.besselj(1, self.x) / Bf.hankelh1(1, self.x)
        elif n == 1:
            new_value = -2 * (1 - self.rho_t) ** 2
            new_value *= Bf.besselj(1, self.x) * Bf.hankelh1(1, self.x_v)
            new_value -= self.mu_1 * self.mu_3
            new_value /= self.mu_4
        else:
            new_value = n * (n + 1) * Bf.besselj(n, self.x)
            new_value *= Bf.hankelh1(n, self.x_v)

            new_value -= self.x * (self.gamma_n(n) * Bf.d1_besselj(n, self.x))
            new_value /= self.xi_n(n)
        return new_value

    def beta_n(self, n: int) -> complex:
        """coefficient :math:`\\beta_n` (3.13) and (3.20)

        :param n: order
        """
        return self._beta_n.item(n)

    def _compute_beta_n(self, n: int) -> complex:
        # computation according to (3.13), (3.15) and (3.21)

        if n == 0:
            new_value = 0.0
        elif n == 1:
            new_value = self.mu_1 * Bf.hankelh1(1, self.x)
            new_value -= self.mu_2 * Bf.besselj(1, self.x)

            new_value *= -(1 - self.rho_t)
            new_value /= self.mu_4
        else:
            new_value = Bf.d1_besselj(n, self.x) * Bf.hankelh1(n, self.x)
            new_value -= Bf.besselj(n, self.x) * Bf.d1_hankelh1(n, self.x)
            new_value *= -self.x
            new_value /= self.xi_n(n)

        return new_value

    @property
    def mu_1(self) -> complex:
        """:math:`\\mu_1` according to (3.16)"""
        return self._mu_1.value

    def _compute_mu_1(self) -> complex:
        out = self.rho_t * Bf.besselj(1, self.x)
        out -= self.x * Bf.d1_besselj(1, self.x)
        return out

    @property
    def mu_2(self) -> complex:
        """:math:`\\mu_2` according to (3.17)"""
        return self._mu_2.value

    def _compute_mu_2(self) -> complex:
        out = self.rho_t * Bf.hankelh1(1, self.x)
        out -= self.x * Bf.d1_hankelh1(1, self.x)
        return out

    @property
    def mu_3(self) -> complex:
        """:math:`\\mu_3` according to (3.18)"""
        return self._mu_3.value

    def _compute_mu_3(self) -> complex:
        out = (1 - 2 * self.rho_t) * Bf.hankelh1(1, self.x_v)
        out += self.x_v * Bf.d1_hankelh1(1, self.x_v)
        return out

    @property
    def mu_4(self) -> complex:
        """:math:`\\mu_4` according to (3.19)"""
        return self._mu_4.value

    def _compute_mu_4(self) -> complex:
        out = 2 * (1 - self.rho_t) ** 2
        out *= Bf.hankelh1(1, self.x)
        out *= Bf.hankelh1(1, self.x_v)
        out += self.mu_2 * self.mu_3
        return out

    # -----------------------------------------------------
    # Methods
    # -----------------------------------------------------

    def particle_velocity(self, t: float) -> complex:
        """Particle velocity

        Returns the value of the particle velocity
        in the direction of the axis of rotational
        symmetry of the radiation field in [m/s]

        :param t: time [s]
        """

        out = Bf.besselj(1, self.x) + self.alpha_n(1) * Bf.hankelh1(1, self.x)
        out -= 2 * self.beta_n(1) * Bf.hankelh1(1, self.x_v)
        out *= self.rho_t * self.A_in(1) * self.k_f
        out /= self.x
        out *= exp(-1j * self.omega * t)
        return out


if __name__ == "__main__":
    pass



from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import Animation

from osaft import (
    FluidScatteringPlot,
    ParticleScatteringPlot,
    ParticleWireframePlot,
)
from osaft.plotting.scattering.base_plotter import BaseScatteringPlotter
from osaft.solutions.base_scattering import BaseScattering
from osaft.tests.plotting.basetest_plotting import BaseTestPlotting


class BaseTestScatteringPlotter(BaseTestPlotting):

    plot_cls: BaseScatteringPlotter

    def setUp(self) -> None:
        super().setUp()
        self.kwargs_options = {"mode": [None, 1, 1], "phase": [0, 1, 2]}
        self.name_prefix = ""

        if not self._is_base_class():
            self.plotting_cls = self._initialize_plotting_class()

    # -------------------------------------------------------------------------
    # Helpers
    # -------------------------------------------------------------------------

    def _generate_kwargs(self) -> dict:
        kwargs = dict()
        # Choose Values
        for key, value in self.kwargs_options.items():
            kwargs[key] = self.rng.choice(value)
        # Remove values from list
        for key, value in kwargs.items():
            self.kwargs_options[key].remove(value)
            if not self.kwargs_options[key]:
                self.kwargs_options.pop(key)
        return kwargs

    def _generate_animation_kwargs(self) -> dict:
        kwargs = self._generate_kwargs()
        try:
            kwargs.pop("phase")
            return kwargs
        except KeyError:
            return kwargs

    def _generate_evolution_kwargs(self) -> dict:
        return self._generate_animation_kwargs()

    @staticmethod
    def _name_from_kwargs(kwargs: dict) -> str:
        name = ""
        for key, value in kwargs.items():
            if isinstance(value, (bool, np.bool_)):
                if value:
                    name = f"{name}_{key}"
            elif value is not None or value != 0:
                name = f"{name}_{key}={value}"
        return name

    def _save_plot(self, fig: plt.Figure, kwargs: dict) -> None:
        name = self.name_prefix + self._name_from_kwargs(kwargs)
        self.save_fig(fig, name)

    @staticmethod
    def _handle_animation(animation: Animation) -> None:
        animation.resume()
        plt.show(block=False)
        plt.pause(1)
        plt.close("all")

    def _is_base_class(self) -> bool:
        return type(self) == BaseTestScatteringPlotter

    # -------------------------------------------------------------------------
    # Abstract Methods
    # -------------------------------------------------------------------------

    def _get_solution(self) -> BaseScattering:
        """Returns the solution class"""

    def _initialize_plotting_class(self) -> BaseScatteringPlotter:
        """Initializes the plotting class"""
        pass

    def _is_valid_field(self, kwargs: dict):
        return True

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        pass

    def fcn_animate(self, **kwargs) -> Animation:
        pass

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        pass

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_plot(self) -> None:
        if self._is_base_class():
            return
        while self.kwargs_options:
            kwargs = self._generate_kwargs()
            if self._is_valid_field(kwargs):
                fig, _ = self.fcn_plot(**kwargs)
                self._save_plot(fig, kwargs)
            else:
                self.assertRaises(ValueError, self.fcn_plot, **kwargs)

    def test_animation(self) -> None:
        if self._is_base_class():
            return
        while self.kwargs_options:
            kwargs = self._generate_animation_kwargs()
            if self._is_valid_field(kwargs):
                anim = self.fcn_animate(**kwargs)
                self._handle_animation(anim)
            else:
                self.assertRaises(ValueError, self.fcn_animate, **kwargs)

    def test_evolution(self) -> None:
        if self._is_base_class():
            return
        while self.kwargs_options:
            kwargs = self._generate_evolution_kwargs()
            if self._is_valid_field(kwargs):
                fig, _ = self.fcn_plot_evolution(**kwargs)
                self._save_plot(fig, kwargs)
            else:
                self.assertRaises(
                    ValueError, self.fcn_plot_evolution, **kwargs
                )


class BaseTestScatterTriPlotter(BaseTestScatteringPlotter):
    def setUp(self) -> None:
        super().setUp()
        self.tri_kwargs_options = {
            "inst": [True, True, False, False, False],
            "symmetric": [True, True, True, False, False, False],
            "tripcolor": [True, True, True, False],
            "cmap": ["jet", "magma"],
            "shading": ["flat", "flat"],
        }
        self.kwargs_options.update(self.tri_kwargs_options)

    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestScatterTriPlotter

    def _generate_animation_kwargs(self) -> dict:
        kwargs = super()._generate_animation_kwargs()
        try:
            kwargs.pop("inst")
            return kwargs
        except KeyError:
            return kwargs

    def _generate_evolution_kwargs(self) -> dict:
        kwargs = super()._generate_evolution_kwargs()
        kwargs_to_remove = ["inst", "cmap", "shading"]
        for kwarg in kwargs_to_remove:
            try:
                kwargs.pop(kwarg)
            except KeyError:
                pass
        return kwargs


class BaseTestFluidScatterPlotter(BaseTestScatterTriPlotter):
    def setUp(self) -> None:
        super().setUp()
        self.fluid_kwargs_options = {
            "scattered": [True, False, False],
            "incident": [True, False, False],
        }
        self.kwargs_options.update(self.fluid_kwargs_options)

    def _initialize_plotting_class(self) -> FluidScatteringPlot:
        sol = self._get_solution()
        r_max = 5 * sol.R_0
        plotting_cls = FluidScatteringPlot(sol, r_max=r_max)
        return plotting_cls

    def _is_valid_field(self, kwargs: dict) -> bool:
        if "scattered" not in kwargs or "incident" not in kwargs:
            return True
        if not kwargs["scattered"] and not kwargs["incident"]:
            return False
        return True

    def test_cmap(self) -> None:
        if self._is_base_class():
            return
        new_cmap = "grays"
        new_div_cmap = "RdYlBu"

        self.plotting_cls.cmap = new_cmap
        self.plotting_cls.div_cmap = new_div_cmap

        self.assertEqual(new_cmap, self.plotting_cls.cmap)
        self.assertEqual(new_div_cmap, self.plotting_cls.div_cmap)


class BaseTestFluidVelocityPlotter(BaseTestFluidScatterPlotter):
    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestFluidVelocityPlotter

    # -------------------------------------------------------------------------
    # Plotting Methods Implementations
    # -------------------------------------------------------------------------

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity(**kwargs)

    def fcn_animate(self, **kwargs) -> Animation:
        return self.plotting_cls.animate_velocity(**kwargs)

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity_evolution(**kwargs)


class BaseTestFluidPressurePlotter(BaseTestFluidScatterPlotter):
    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestFluidPressurePlotter

    # -------------------------------------------------------------------------
    # Plotting Methods Implementations
    # -------------------------------------------------------------------------

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_pressure(**kwargs)

    def fcn_animate(self, **kwargs) -> Animation:
        return self.plotting_cls.animate_pressure(**kwargs)

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_pressure_evolution(**kwargs)


class BaseTestFluidPotentialPlotter(BaseTestFluidScatterPlotter):
    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestFluidPotentialPlotter

    # -------------------------------------------------------------------------
    # Plotting Methods Implementations
    # -------------------------------------------------------------------------

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity_potential(**kwargs)

    def fcn_animate(self, **kwargs) -> Animation:
        return self.plotting_cls.animate_velocity_potential(**kwargs)

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity_potential_evolution(**kwargs)


class BaseTestFluidQuiverPlot(BaseTestFluidScatterPlotter):
    def setUp(self) -> None:
        super().setUp()
        # Add additional kwargs
        self.quiver_kwargs = {"quiver_color": [None, "black", "white"]}
        self.kwargs_options.update(self.quiver_kwargs)

    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestFluidQuiverPlot

    # -------------------------------------------------------------------------
    # Plotting Methods Implementations
    # -------------------------------------------------------------------------

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity(**kwargs)

    def fcn_animate(self, **kwargs) -> Animation:
        return self.plotting_cls.animate_velocity(**kwargs)

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity_evolution(**kwargs)


class BaseTestParticleScatterPlotter(BaseTestScatterTriPlotter):
    def setUp(self) -> None:
        super().setUp()
        # Add additional kwargs
        self.particle_kwargs_options = {
            "displacement": [True, False],
        }
        self.kwargs_options.update(self.particle_kwargs_options)

    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestParticleScatterPlotter

    def _initialize_plotting_class(self) -> ParticleScatteringPlot:
        plotting_cls = ParticleScatteringPlot(self._get_solution())
        return plotting_cls

    # -------------------------------------------------------------------------
    # Plotting Methods Implementations
    # -------------------------------------------------------------------------

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity(**kwargs)

    def fcn_animate(self, **kwargs) -> Animation:
        return self.plotting_cls.animate_velocity(**kwargs)

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_velocity_evolution(**kwargs)


class BaseTestParticleQuiverPlot(BaseTestParticleScatterPlotter):
    def setUp(self) -> None:
        super().setUp()
        # Add additional kwargs
        self.quiver_kwargs = {"quiver_color": [None, "black", "white"]}
        self.kwargs_options.update(self.quiver_kwargs)

    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestParticleQuiverPlot


class BaseTestWireFrameScatterPlotter(BaseTestScatteringPlotter):
    def _is_base_class(self) -> bool:
        if super()._is_base_class():
            return True
        return type(self) == BaseTestWireFrameScatterPlotter

    def _initialize_plotting_class(self):
        plotting_cls = ParticleWireframePlot(self._get_solution())
        return plotting_cls

    # -------------------------------------------------------------------------
    # Plotting Methods Implementations
    # -------------------------------------------------------------------------

    def fcn_plot(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot(**kwargs)

    def fcn_animate(self, **kwargs) -> Animation:
        return self.plotting_cls.animate(**kwargs)

    def fcn_plot_evolution(self, **kwargs) -> tuple[plt.Figure, plt.Axes]:
        return self.plotting_cls.plot_evolution(**kwargs)


if __name__ == "__main__":
    pass
